"""Translates a set of VCF files by adding standardized tags.

Reads incoming VCF files determining appropriate "origin caller" (e.g. MuTect);
emits a new translated file. The translated file is similar to the input
file, with these exceptions:
    * translate will add a filter flag anomalous VCF records, i.e. records
        that don't conform to the standard; for example both Strelka and
        Varscan emit VCF records with invalid ALT values.
    * translate will add new Jacquard-standard FORMAT tags that augment the
        caller specific tags (e.g. a Varscan FREQ tag would generates a new
        JQ_VS_AF tag).

There will typically be a translated VCF file for each input VCF file.
Unrecognized VCFs are not copied to output.
The origin caller of a VCFs is recognized in part by the content of metaheaders,
so it's imperative that VCF metaheaders be present and accurate.
"""
from __future__ import print_function, absolute_import, division

import argparse
from collections import defaultdict
import glob
import os

import jacquard.utils.logger as logger
import jacquard.utils.utils as utils
from jacquard.variant_caller_transforms import variant_caller_factory
from jacquard.utils.vcf import FileReader, FileWriter


_FILE_OUTPUT_SUFFIX = "translatedTags"


class _ExcludeMalformedRef(object):
    #pylint: disable=too-few-public-methods
    _VALID_REF = set(list("ACGTNacgtn"))
    _TAG_ID = "JQ_EXCLUDE_MALFORMED_REF"

    def __init__(self):
        self.metaheader = ('##FILTER=<'
                           'ID={},'
                           #pylint: disable=line-too-long
                           'Description="The format of the reference value for this variant record does not comply with VCF standard.">')\
                           .format(self._TAG_ID)

    def _is_valid_ref(self, record):
        return set(list(record.ref)).issubset(self._VALID_REF)

    def add_tag_values(self, record):
        if not self._is_valid_ref(record):
            record.add_or_replace_filter(self._TAG_ID)

class _ExcludeMalformedAlt(object):
    #pylint: disable=too-few-public-methods
    _VALID_ALT = set(list("*.ACGTNacgtn,"))
    _TAG_ID = "JQ_EXCLUDE_MALFORMED_ALT"

    def __init__(self):
        self.metaheader = ('##FILTER=<'
                           'ID={},'
                           #pylint: disable=line-too-long
                           'Description="The the format of the alternate allele value for this variant record does not comply with VCF standard.">')\
                           .format(self._TAG_ID)

    def _is_valid_alt(self, record):
        valid_characters = set(list(record.alt)).issubset(self._VALID_ALT)
        invalid_characters = set(list("*.")).issubset(set(list(record.alt)))

        return valid_characters and not invalid_characters

    def add_tag_values(self, record):
        if not self._is_valid_alt(record):
            record.add_or_replace_filter(self._TAG_ID)

class _ExcludeMissingAlt(object):
    #pylint: disable=too-few-public-methods
    _MISSING_ALT = "."
    _TAG_ID = "JQ_EXCLUDE_MISSING_ALT"

    def __init__(self):
        self.metaheader = ('##FILTER='
                           '<ID={},'
                           #pylint: disable=line-too-long
                           'Description="The alternate allele is missing for this variant record.">')\
                           .format(self._TAG_ID)

    def add_tag_values(self, record):
        if record.alt == self._MISSING_ALT:
            record.add_or_replace_filter(self._TAG_ID)


def _build_file_readers(input_dir):
    in_files = glob.glob(os.path.join(input_dir, "*"))
    file_readers = []

    for in_file in in_files:
        file_readers.append(FileReader(in_file))

    return file_readers

def _mangle_output_filename(input_file):
    basename, extension = os.path.splitext(os.path.basename(input_file))
    return ".".join([basename, _FILE_OUTPUT_SUFFIX, extension.strip(".")])

def _check_snp_indel_pairings(altered_file_names, args):
    if not set([len(i) for i in altered_file_names.values()]) == set([1]):
        if not args.allow_inconsistent_sample_sets:
            error = 0
            for file_names in altered_file_names.values():
                if len(file_names) % 2 != 0:
                    message = ("File {} was missing a "
                               "corresponding snp/indel file.")
                    logger.error(message, file_names)
                    error = 1
            if error:
                message = ("Some VCFs were missing either a snp/snvs "
                            "or an indel/indels file. Review "
                            "inputs/command options to align file "
                            "pairings or use the flag "
                            "--allow_inconsistent_sample_sets.")
                raise utils.UsageError(message)

def _validate_file_pairings(args, claimed_vcf_readers):
    input_path = args.input
    if not os.path.isdir(input_path):
        input_path = os.path.dirname(input_path)

    input_vcfs = sorted(glob.glob(os.path.join(input_path, "*.vcf")))
    altered_file_names = defaultdict(list)
    for vcf_reader in claimed_vcf_readers:
        for input_vcf in input_vcfs:
            basename = os.path.basename(input_vcf)
            word_list = vcf_reader.expected_file_format()
            file_names = [i for i in basename.split(".") if i not in word_list]
            joined_file_names = ".".join(file_names)
            if len(joined_file_names) != len(basename):
                altered_file_names[joined_file_names].append(basename)

    _check_snp_indel_pairings(altered_file_names, args)

def validate_args(args):
    unclaimed_readers, trans_vcf_readers = _claim_readers(args)
    if unclaimed_readers and not args.force:
        raise utils.UsageError(_build_validation_message(unclaimed_readers))
    elif not trans_vcf_readers:
        raise utils.UsageError(("no vcfs in input dir "
                                "[{}] can be translated.").format(args.input))
    _validate_file_pairings(args, trans_vcf_readers)

def _build_validation_message(unclaimed_readers):
    total_unclaimed = len(unclaimed_readers)
    if total_unclaimed == 1:
        unclaimed_details = "file [{}]".format(unclaimed_readers[0].file_name)
    else:
        cutoff = 5
        file_names = [reader.file_name for reader in unclaimed_readers]
        file_names = file_names[0:min(cutoff, total_unclaimed)]
        unclaimed_list = ", ".join(file_names)
        if total_unclaimed > cutoff:
            omitted = total_unclaimed - cutoff
            unclaimed_list += ", ...({} file(s) omitted)".format(omitted)
        unclaimed_details = "files [{}]".format(unclaimed_list)

    return ("{} input {} cannot be "
            "translated; review input dir and try again, or "
            "use '--force' to ignore these files.").format(total_unclaimed,
                                                           unclaimed_details)

def _claim_readers(args):
    input_dir = os.path.abspath(args.input)
    file_readers = _build_file_readers(input_dir)
    factory = variant_caller_factory.VariantCallerFactory(args)

    return factory.claim(file_readers)

def _log_unclaimed_readers(unclaimed_readers):
    unclaimed_log_message = "The input file [{}] will not be translated"
    for reader in sorted(unclaimed_readers):
        msg = unclaimed_log_message.format(reader.file_name)
        logger.warning(msg)

def _translate_files(trans_vcf_reader,
                     new_tags,
                     execution_context,
                     file_writer):
    try:
        trans_vcf_reader.open()
        file_writer.open()

        _write_headers(trans_vcf_reader,
                       new_tags,
                       execution_context,
                       file_writer)

        for record in trans_vcf_reader.vcf_records():
            for tag in new_tags:
                tag.add_tag_values(record)
            file_writer.write(record.text())

    finally:
        trans_vcf_reader.close()
        file_writer.close()

def _write_headers(reader, new_tags, execution_context, file_writer):
    headers = reader.metaheaders
    headers.extend(execution_context)
    for tag in new_tags:
        headers.append(tag.metaheader)

    sorted_headers = utils.sort_metaheaders(headers)
    sorted_headers.append(reader.column_header)

    file_writer.write("\n".join(sorted_headers) + "\n")

def get_required_input_output_types():
    return ("directory", "directory")

def report_prediction(args):
    input_dir = os.path.abspath(args.input)
    file_readers = _build_file_readers(input_dir)
    output_file_names = set()

    for reader in file_readers:
        mangled_fname = _mangle_output_filename(reader.file_name)
        extension = os.path.splitext(os.path.basename(mangled_fname))[1]
        if extension == ".vcf":
            output_file_names.add(mangled_fname)

    return output_file_names

def add_subparser(subparser):
    #pylint: disable=line-too-long
    parser = subparser.add_parser("translate", formatter_class=argparse.RawTextHelpFormatter, help="Accepts a directory of VCF results (and VarScan high confidence files). Creates a new directory of VCFs, adding Jacquard-specific FORMAT tags for each VCF record.")
    parser.add_argument("input", help="Path to directory containing VCFs (and VarScan high confidence files). Other file types ignored")
    parser.add_argument("output", help="Path to Jacquard-tagged VCFs. Will create if doesn't exist and will overwrite files in output directory as necessary")
    parser.add_argument("-v", "--verbose", action='store_true')
    parser.add_argument("--force", action='store_true', help="Overwrite contents of output directory")
    parser.add_argument("--varscan_hc_filter_file_regex",
                        help=("Regex pattern that identifies optional VarScan high-confidence filter files.\n"
                              "The VCF, high-confidence file pairs should share the same prefix.\n"
                              "For example, given patientA.snp.vcf, patientA.indel.vcf, patientA.snp.fpfilter.pass, patientA.indel.fpfilter.pass,\n"
                              "you could enable this option as varscan_hc_filter_file_regex='.fpfilter.pass$'"))
    parser.add_argument("--allow_inconsistent_sample_sets", help="Allow inconsistent sample sets to be used")
    parser.add_argument("--log_file", help="Path to log file destination. Defaults to current working directory if not specified.")

#TODO (cgates): This module is both a command and also manipulates VcfRecords
# like a caller. This is the only body of code that does both these things.
# Does this bother anyone else?
def execute(args, execution_context):
    validate_args(args)

    output_dir = os.path.abspath(args.output)
    unclaimed_readers, trans_vcf_readers = _claim_readers(args)

    _log_unclaimed_readers(unclaimed_readers)

    logger.info("Processing [{}] VCF file(s) from [{}]",
                len(trans_vcf_readers),
                args.input)

    new_tags = [_ExcludeMalformedRef(),
                _ExcludeMalformedAlt(),
                _ExcludeMissingAlt()]

    for i, trans_vcf_reader in enumerate(trans_vcf_readers):
        logger.info("Translating file {}/{} [{}]",
                    i + 1,
                    len(trans_vcf_readers),
                    trans_vcf_reader.file_name)
        new_filename = _mangle_output_filename(trans_vcf_reader.file_name)
        output_filepath = os.path.join(output_dir, new_filename)
        file_writer = FileWriter(output_filepath)
        _translate_files(trans_vcf_reader,
                         new_tags,
                         execution_context,
                         file_writer,)

    logger.info("Wrote [{}] VCF file(s)", len(trans_vcf_readers))



"""Classes to summarize data for a sample-variant or variant as a whole.

A collection of individual Tag classes hold the metaheader and logic to
transform Jacquard-standardized VcfRecords.
"""
from __future__ import print_function, absolute_import, division

from decimal import Decimal
import re

import jacquard.utils.utils as utils
import jacquard.variant_caller_transforms.common_tags as common_tags


JQ_SUMMARY_TAG = "JQ_SUMMARY_"
JQ_REPORTED = "CALLERS_REPORTED_COUNT"
JQ_REPORTED_LIST = "CALLERS_REPORTED_LIST"
JQ_SAMPLES_REPORTED = "SAMPLES_REPORTED_COUNT"
JQ_PASSED = "CALLERS_PASSED_COUNT"
JQ_PASSED_LIST = "CALLERS_PASSED_LIST"
JQ_SAMPLES_PASSED = "SAMPLES_PASSED_COUNT"

def _aggregate_numeric_values(values, function):
    split_values = [x.split(",") for x in values]
    lengths = set([len(x) for x in split_values])
    if len(lengths) > 1:
        return 0

    #pylint: disable=star-args
    transposed_values = [list(x) for x in zip(*split_values)]
    string_values = []
    for group in transposed_values:
        string_values.append(function([Decimal(i) for i in group]))

    return ",".join(string_values)

def _get_non_null_values(record, sample, tag_name_regex):
    values = []
    try:
        for tag, value in list(record.sample_tag_values[sample].items()):
            if tag_name_regex.match(tag) and value != ".":
                values.append(value)
    except KeyError:
        raise utils.JQException("Sample [{}] was not recognized".format(sample))

    return values

def _average(numeric_values):
    average = str(sum(numeric_values)/Decimal(len(numeric_values)))
    return utils.round_two_digits(average)

def _range(numeric_values):
    if len(numeric_values) > 1:
        value_range = str(max(numeric_values) - min(numeric_values))
        return utils.round_two_digits(value_range)
    return "."

def _count(numeric_values):
    return str(sum(i == 1 for i in numeric_values))

def _build_new_tags(vcf_record, tags, sample):
    desired_tags = []
    for tag in tags:
        desired_tag = vcf_record.sample_tag_values[sample][tag].split(",")

        #don't include null values in avg calculation
        altered_tag = [x for x in desired_tag if x != "."]
        if len(altered_tag) != 0:
            desired_tags.append(altered_tag)

    return desired_tags

def _add_caller_list_values(pattern, vcf_record, jq_global_variable):
    sample_tag = {}
    for sample, tags in vcf_record.sample_tag_values.items():
        sample_tag[sample] = []
        for tag in tags:
            pattern_match = pattern.match(tag)
            if pattern_match and tags[tag] != "0" and tags[tag] != ".":
                sample_tag[sample].append(pattern_match.group(1))

    for sample in sample_tag:
        sample_tag[sample] = ",".join(sorted(sample_tag[sample]))
        if not sample_tag[sample]:
            sample_tag[sample] = "."
    vcf_record.add_sample_tag_value(JQ_SUMMARY_TAG + jq_global_variable,
                                    sample_tag)

def _add_caller_count_values(pattern, vcf_record, jq_global_variable):
    sample_tag = {}
    for sample, tags in vcf_record.sample_tag_values.items():
        sample_tag[sample] = 0
        for tag in tags:
            if pattern.match(tag) and tags[tag] != "0" and tags[tag] != ".":
                sample_tag[sample] += int(tags[tag])
    vcf_record.add_sample_tag_value(JQ_SUMMARY_TAG + jq_global_variable,
                                    sample_tag)

def _add_sample_count_values(vcf_record,
                             jq_callers_global_variable,
                             jq_samples_global_variable):
    count = 0
    for tags in vcf_record.sample_tag_values.values():
        tag_key = JQ_SUMMARY_TAG + jq_callers_global_variable
        if tag_key in tags and tags[tag_key] != "0" and tags[tag_key] != ".":
            count += 1

    info_key = JQ_SUMMARY_TAG + jq_samples_global_variable
    vcf_record.add_info_field("=".join([info_key, str(count)]))

class _CallersReportedListTag(object):
    #pylint: disable=too-few-public-methods
    def __init__(self):
        self.metaheader = self._get_metaheader()
        self.pattern = re.compile(r"JQ_(.*?)_{}"\
                                  .format(common_tags.CALLER_REPORTED_TAG))

    @staticmethod
    def _get_metaheader():
        return ('##FORMAT=<ID={}{},'
                'Number=.,'
                'Type=String,'
                #pylint: disable=line-too-long
                'Description="Comma-separated list variant callers which listed this variant in the Jacquard tagged VCF">')\
                .format(JQ_SUMMARY_TAG,
                        JQ_REPORTED_LIST)

    def add_tag_values(self, vcf_record):
        _add_caller_list_values(self.pattern, vcf_record, JQ_REPORTED_LIST)

class _CallersReportedTag(object):
    #pylint: disable=too-few-public-methods
    def __init__(self):
        self.metaheader = self._get_metaheader()
        self.pattern = re.compile(r"JQ_(.*?)_{}"\
                                  .format(common_tags.CALLER_REPORTED_TAG))

    @staticmethod
    def _get_metaheader():
        return ('##FORMAT=<ID={}{},'
                'Number=1,'
                'Type=Integer,'
                #pylint: disable=line-too-long
                'Description="Count of variant callers which listed this variant in the Jacquard tagged VCF">')\
                .format(JQ_SUMMARY_TAG,
                        JQ_REPORTED)

    def add_tag_values(self, vcf_record):
        _add_caller_count_values(self.pattern, vcf_record, JQ_REPORTED)

class _CallersPassedListTag(object):
    #pylint: disable=too-few-public-methods
    def __init__(self):
        self.metaheader = self._get_metaheader()
        self.pattern = re.compile(r"JQ_(.*?)_{}"\
                                  .format(common_tags.CALLER_PASSED_TAG))

    @staticmethod
    def _get_metaheader():
        return ('##FORMAT=<ID={}{},'
                'Number=.,'
                'Type=String,'
                #pylint: disable=line-too-long
                'Description="Comma-separated list of variant caller short-names where FILTER = PASS for this variant in the Jacquard tagged VCF">')\
                .format(JQ_SUMMARY_TAG,
                        JQ_PASSED_LIST)

    def add_tag_values(self, vcf_record):
        _add_caller_list_values(self.pattern, vcf_record, JQ_PASSED_LIST)

class _CallersPassedTag(object):
    #pylint: disable=too-few-public-methods
    def __init__(self):
        self.metaheader = self._get_metaheader()
        self.pattern = re.compile(r"JQ_(.*?)_{}"\
                                  .format(common_tags.CALLER_PASSED_TAG))

    @staticmethod
    def _get_metaheader():
        return ('##FORMAT=<ID={}{},'
                'Number=1,'
                'Type=Integer,'
                #pylint: disable=line-too-long
                'Description="Count of variant callers where FILTER = PASS for this variant in the Jacquard tagged VCF">')\
                .format(JQ_SUMMARY_TAG,
                        JQ_PASSED)

    def add_tag_values(self, vcf_record):
        _add_caller_count_values(self.pattern, vcf_record, JQ_PASSED)

class _SamplesReported(object):
    #pylint: disable=too-few-public-methods
    def __init__(self):
        self.metaheader = self._get_metaheader()

    @staticmethod
    def _get_metaheader():
        return ('##INFO=<ID={}{},'
                'Number=1,'
                'Type=Integer,'
                #pylint: disable=line-too-long
                'Description="Count of samples where this variant appeared in any of the Jacquard tagged VCFs (regardless of quality/filtering)">')\
                .format(JQ_SUMMARY_TAG,
                        JQ_SAMPLES_REPORTED)

    @staticmethod
    def add_tag_values(vcf_record):
        _add_sample_count_values(vcf_record, JQ_REPORTED, JQ_SAMPLES_REPORTED)

class _SamplesPassed(object):
    #pylint: disable=too-few-public-methods
    def __init__(self):
        self.metaheader = self._get_metaheader()

    @staticmethod
    def _get_metaheader():
        return ('##INFO=<ID={}{},'
                'Number=1,'
                'Type=Integer,'
                #pylint: disable=line-too-long
                'Description="Count of samples where a variant caller passed the filter in any of the Jacquard tagged VCFs">')\
                .format(JQ_SUMMARY_TAG,
                        JQ_SAMPLES_PASSED)

    @staticmethod
    def add_tag_values(vcf_record):
        _add_sample_count_values(vcf_record, JQ_PASSED, JQ_SAMPLES_PASSED)

class _AlleleFreqRangeTag(object):
    #pylint: disable=too-few-public-methods
    _TAG_ID = "{}AF_RANGE".format(JQ_SUMMARY_TAG)
    _PATTERN = re.compile("^JQ_.*_AF$")

    def __init__(self):
        self.metaheader = self._get_metaheader()

    @staticmethod
    def _get_metaheader():
        return ('##FORMAT=<ID={},'
                'Number=1,'
                'Type=Float,'
                ##pylint: disable=line-too-long
                'Description="Max(allele frequency) - min (allele frequency) across recognized callers.">')\
                .format(_AlleleFreqRangeTag._TAG_ID)

    @staticmethod
    def add_tag_values(record):
        new_sample_tag_values = {}
        for sample in record.sample_tag_values:
            tag_values = _get_non_null_values(record,
                                              sample,
                                              _AlleleFreqRangeTag._PATTERN)

            aggregated_values = "."
            if tag_values:
                aggregated_values = _aggregate_numeric_values(tag_values,
                                                              _range)
            if not aggregated_values:
                #pylint: disable=line-too-long
                raise utils.JQException("Error summarizing values {} at record [{}:{} {}]",
                                        list(tag_values),
                                        record.chrom,
                                        record.pos,
                                        sample)

            new_sample_tag_values[sample] = aggregated_values

        record.add_sample_tag_value(_AlleleFreqRangeTag._TAG_ID,
                                    new_sample_tag_values)

class _AlleleFreqAverageTag(object):
    #pylint: disable=too-few-public-methods
    _TAG_ID = "{}AF_AVERAGE".format(JQ_SUMMARY_TAG)
    _PATTERN = re.compile("^JQ_.*_AF$")

    def __init__(self):
        self.metaheader = self._get_metaheader()

    @staticmethod
    def _get_metaheader():
        return ('##FORMAT=<ID={0},'
                'Number=1,'
                'Type=Float,'
                ##pylint: disable=line-too-long
                'Description="Average allele frequency across recognized variant callers that reported frequency for this position [average(JQ_*_AF)].">')\
                .format(_AlleleFreqAverageTag._TAG_ID)

    @staticmethod
    def add_tag_values(record):
        new_sample_tag_values = {}
        for sample in record.sample_tag_values:
            tag_values = _get_non_null_values(record,
                                              sample,
                                              _AlleleFreqAverageTag._PATTERN)

            aggregated_values = "."
            if tag_values:
                aggregated_values = _aggregate_numeric_values(tag_values,
                                                              _average)
            if not aggregated_values:
                #pylint: disable=line-too-long
                raise utils.JQException("Error summarizing values {} at record [{}:{} {}]",
                                        list(tag_values),
                                        record.chrom,
                                        record.pos,
                                        sample)

            new_sample_tag_values[sample] = aggregated_values

        record.add_sample_tag_value(_AlleleFreqAverageTag._TAG_ID,
                                    new_sample_tag_values)

class _DepthRangeTag(object):
    #pylint: disable=too-few-public-methods
    _TAG_ID = "{}DP_RANGE".format(JQ_SUMMARY_TAG)
    _PATTERN = re.compile("^JQ_.*_DP$")

    def __init__(self):
        self.metaheader = self._get_metaheader()

    @staticmethod
    def _get_metaheader():
        return ('##FORMAT=<ID={},'
                'Number=1,'
                'Type=Float,'
                ##pylint: disable=line-too-long
                'Description="Max(depth) - min (depth) across recognized callers.">'
                .format(_DepthRangeTag._TAG_ID))

    @staticmethod
    def add_tag_values(record):
        new_sample_tag_values = {}
        for sample in record.sample_tag_values:
            tag_values = _get_non_null_values(record,
                                              sample,
                                              _DepthRangeTag._PATTERN)

            aggregated_values = "."
            if tag_values:
                aggregated_values = _aggregate_numeric_values(tag_values,
                                                              _range)
            if not aggregated_values:
                #pylint: disable=line-too-long
                raise utils.JQException("Error summarizing values {} at record [{}:{} {}]",
                                        list(tag_values),
                                        record.chrom,
                                        record.pos,
                                        sample)

            new_sample_tag_values[sample] = aggregated_values

        record.add_sample_tag_value(_DepthRangeTag._TAG_ID,
                                    new_sample_tag_values)

class _DepthAverageTag(object):
    #pylint: disable=too-few-public-methods
    _TAG_ID = "{}DP_AVERAGE".format(JQ_SUMMARY_TAG)
    _PATTERN = re.compile("^JQ_.*_DP$")

    def __init__(self):
        self.metaheader = self._get_metaheader()

    @staticmethod
    def _get_metaheader():
        return ('##FORMAT=<ID={},'
                'Number=1,'
                'Type=Float,'
                ##pylint: disable=line-too-long
                'Description="Average allele frequency across recognized variant callers that reported frequency for this position; rounded to integer [round(average(JQ_*_DP))].">'
                .format(_DepthAverageTag._TAG_ID))

    @staticmethod
    def add_tag_values(record):
        new_sample_tag_values = {}
        for sample in record.sample_tag_values:
            tag_values = _get_non_null_values(record,
                                              sample,
                                              _DepthAverageTag._PATTERN)

            aggregated_values = "."
            if tag_values:
                aggregated_values = _aggregate_numeric_values(tag_values,
                                                              _average)
            if not aggregated_values:
                #pylint: disable=line-too-long
                raise utils.JQException("Error summarizing values {} at record [{}:{} {}]",
                                        list(tag_values),
                                        record.chrom,
                                        record.pos,
                                        sample)

            new_sample_tag_values[sample] = aggregated_values

        record.add_sample_tag_value(_DepthAverageTag._TAG_ID,
                                    new_sample_tag_values)

class _SomaticTag(object):
    #pylint: disable=too-few-public-methods
    _TAG_ID = "{}SOM_COUNT".format(JQ_SUMMARY_TAG)
    _PATTERN = re.compile("^JQ_.*_HC_SOM$")

    def __init__(self):
        self.metaheader = self._get_metaheader()

    @staticmethod
    def _get_metaheader():
        som_count = ('##FORMAT=<ID={},'
                     'Number=1,'
                     'Type=Integer,'
                     ##pylint: disable=line-too-long
                     'Description="Count of recognized variant callers that reported confident somatic call for this sample-position.">')\
                     .format(_SomaticTag._TAG_ID)
        return som_count

    @staticmethod
    def add_tag_values(record):
        new_sample_tag_values = {}
        for sample in record.sample_tag_values:
            tag_values = _get_non_null_values(record,
                                              sample,
                                              _SomaticTag._PATTERN)
            aggregated_values = "."
            if tag_values:
                aggregated_values = _aggregate_numeric_values(tag_values,
                                                              _count)
            if not aggregated_values:
                #pylint: disable=line-too-long
                raise utils.JQException("Error summarizing values {} at record [{}:{} {}]",
                                        list(tag_values),
                                        record.chrom,
                                        record.pos,
                                        sample)

            new_sample_tag_values[sample] = aggregated_values

        record.add_sample_tag_value(_SomaticTag._TAG_ID,
                                    new_sample_tag_values)

class SummarizeCaller(object):
    """Provides metaheaders for VcfReader; adds summary tags to VcfRecords."""
    def __init__(self):
        self.tags = [_CallersReportedTag(),
                     _CallersPassedTag(),
                     _CallersReportedListTag(),
                     _CallersPassedListTag(),
                     _SamplesReported(),
                     _SamplesPassed(),
                     _AlleleFreqAverageTag(),
                     _AlleleFreqRangeTag(),
                     _DepthAverageTag(),
                     _DepthRangeTag(),
                     _SomaticTag()]

    def add_tags(self, vcf_record):
        for tag in self.tags:
            tag.add_tag_values(vcf_record)
        return vcf_record

    def get_metaheaders(self):
        return [tag.metaheader for tag in self.tags]



#!/usr/bin/env python
"""Entry-point for suite of sub-commands.

The only executable module in the project; this module
 * validates command line args
 * manages use of temp directories (to keep output clean and atomic)
 * dipatches to sub-commands as appropriate
 * attempts to deal with usage and run-time errors

Jacquard first writes results to temp dir and only copies results on successful
    completion.

Then architecture of Jacquard modules can be divided into:
 * commands : These transform files or directories (e.g. translate.py) and are
     indirectly executable through the jacquard module. Each command must
     implement an execute method that does the heavy lifting along with some
     simpler methods that expedite command validation
 * callers : These transform VcfRecords (e.g. mutect.py). They typically have
     a collection of tag classes; where each tag holds the metaheader and
     code to transform a single VcfRecord. Note that a caller could
     manipulate any aspect of a VcfRecord, but (by strong convention) typically
     only adds information, for example add a sample-format tag, add an info
     field, or add a filter field.
 * helpers : Common functionality (e.g. command_validator, vcf, logger, etc.)
"""
##   Copyright 2014 Bioinformatics Core, University of Michigan
##
##   Licensed under the Apache License, Version 2.0 (the "License");
##   you may not use this file except in compliance with the License.
##   You may obtain a copy of the License at
##
##       http://www.apache.org/licenses/LICENSE-2.0
##
##   Unless required by applicable law or agreed to in writing, software
##   distributed under the License is distributed on an "AS IS" BASIS,
##   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
##   See the License for the specific language governing permissions and
##   limitations under the License.
from __future__ import print_function, absolute_import, division

import argparse
from datetime import datetime
import distutils.dir_util
import os
import re
import shutil
import signal
import sys
import traceback

from jacquard import __version__
import jacquard.utils.command_validator as command_validator
import jacquard.expand as expand
import jacquard.utils.logger as logger
import jacquard.merge as merge
import jacquard.summarize as summarize
import jacquard.translate as translate
import jacquard.utils.utils as utils
from jacquard.variant_caller_transforms import variant_caller_factory


_SUBCOMMANDS = [translate,
                merge,
                summarize,
                expand]


class _JacquardArgumentParser(argparse.ArgumentParser):
    """Argument parser that raises UsageError instead of exiting."""
    #pylint: disable=too-few-public-methods

    @staticmethod
    def _remessage_invalid_subparser(message):
        pattern = r"argument subparser_name: invalid choice: '(.*?)' \((.*)\)"
        match = re.match(pattern, message)
        if match:
            message = ("'{}' is not a Jacquard command;"
                       " {}.").format(match.group(1), match.group(2))
        return message

    def error(self, message):
        '''Suppress default exit behavior'''
        message = self._remessage_invalid_subparser(message)
        raise utils.UsageError(message)


def _cleanup(temp_working_dir):
    if temp_working_dir and os.path.exists(temp_working_dir):
        logger.debug("Cleaning up tmp directory")
        shutil.rmtree(temp_working_dir)


def _move_tmp_contents_to_original(args):
    temp_dir = args.temp_working_dir
    dest_dir = os.path.dirname(args.temp_working_dir)

    logger.debug("Moving files from tmp directory [{}] to output "
                 "directory [{}]",
                 temp_dir,
                 dest_dir)
    distutils.dir_util.copy_tree(temp_dir, dest_dir)
    logger.debug("Output saved to [{}]", os.path.abspath(args.original_output))
    logger.info("Output saved to [{}]", args.original_output)
    _cleanup(temp_dir)


def _parse_command_line_args(modules, arguments):
    parser = _JacquardArgumentParser(
        usage="jacquard",
#         formatter_class=argparse.RawDescriptionHelpFormatter,
        formatter_class=argparse.RawTextHelpFormatter,
        # pylint: disable=line-too-long
        description='''type 'jacquard <command> -h' for help on a specific command''',
        epilog="See https://github.com/umich-brcf-bioinf/Jacquard for more info.")

    parser.add_argument("-V",
                        "--version",
                        action='version',
                        version=_version_text())
    subparsers = parser.add_subparsers(title="subcommands",
                                       dest="subparser_name")

    module_dispatch = {}
    for module in modules:
        module.add_subparser(subparsers)
        short_name = module.__name__.split('.')[-1]
        module_dispatch[short_name] = module

    args = parser.parse_args(arguments)
    return module_dispatch[args.subparser_name], args


def _set_interrupt_handler(target=signal.signal):
    def _handler(dummy_signum, dummy_frame):
        msg = "WARNING: Jacquard was interrupted before completing."
        try:
            logger.debug(msg)
        finally:
            print(msg, file=sys.stderr)
            exit(1)
    target(signal.SIGINT, _handler)
    target(signal.SIGTERM, _handler)


def _version_text():
    callers = variant_caller_factory.SUPPORTED_CALLER_VERSIONS.items()
    caller_versions = [key + " " + value for key, value in callers]
    caller_version_string = "\n\t".join(caller_versions)
    return "Jacquard v{0}\nSupported variant callers:\n\t{1}".\
        format(__version__, caller_version_string)

def _get_execution_context(command):
    cwd = os.path.dirname(os.getcwd())
    now = datetime.now().strftime('%Y-%m-%d %H:%M:%S')

    return ['##jacquard=<Timestamp="{}",Command="{}",Cwd="{}">'.format(now,
                                                                       command,
                                                                       cwd)]

def _dispatch(modules, arguments):
    try:
        command, args = _parse_command_line_args(modules, arguments)
        execution_context = _get_execution_context(command)

        logger.initialize_logger(args)
        logger.debug("Jacquard run begins")
        logger.debug("cwd|{}", os.getcwd())
        logger.debug("command|{}", " ".join(arguments))

        command_validator.preflight(command, args)

        logger.info("Jacquard begins (v{})", __version__)
        logger.info("Saving log to [{}]",
                    os.path.basename(logger.log_filename))
        logger.debug("Writing output to tmp directory [{}]",
                     args.temp_working_dir)

        command.execute(args, execution_context)

        _move_tmp_contents_to_original(args)

        if logger.WARNING_OCCURRED:
            logger.info("Done. (See warnings above)")
        else:
            logger.info("Done")

    except utils.UsageError as usage_error:
        message = "Jacquard usage problem: {}".format(str(usage_error))
        logger.debug(message)
        print(message, file=sys.stderr)
        try:
            print("See 'jacquard {} --help'.".format(args.subparser_name),
                  file=sys.stderr)
        except NameError: #could not determine the command
            print("See 'jacquard --help'.", file=sys.stderr)
        sys.exit(1)
    except Exception as exception: #pylint: disable=broad-except
        logger.error(str(exception))
        logger.error("Jacquard encountered an unanticipated problem. "
                     "Please review log file and contact your sysadmin "
                     "or Jacquard support for assistance.")
        logger.debug(traceback.format_exc())
        sys.exit(1)
    finally:
        try:
            _cleanup(args.temp_working_dir)
        except Exception: #pylint: disable=broad-except
            pass #we tried


def main():
    _set_interrupt_handler()
    _dispatch(_SUBCOMMANDS, sys.argv[1:])


if __name__ == '__main__':
    main()



"""Expands a VCF file to tab-separated text file

The number of rows will be the same as the number of records in the VCF (i.e.
line count of VCF including the column header - number of metaheaders).

Will create a column for each:
 * static field (e.g. CHROM, POS, ID, REF, ALT, QUAL, FILTER)
 * INFO field listed in the metaheader
 * combination of sample name and FORMAT tag listed in the metaheader

Expand relies on accurate metaheaders; it will not expand any fields absent
from the metaheaders.
"""
from __future__ import print_function, absolute_import, division

import argparse
from collections import OrderedDict
import os
import re

import jacquard.utils.logger as logger
import jacquard.utils.utils as utils
import jacquard.utils.vcf as vcf


UNUSED_REGEX_WARNING_FORMAT = ("The expression [{}] in column specification "
                               "file [{}:{}] didn't match any input columns; "
                               "columns may have matched earlier expressions, "
                               "or this expression may be irrelevant.")

def _read_col_spec(col_spec):
    spec_file = open(col_spec, "r")
    columns = []
    for line in spec_file:
        columns.append(line.rstrip())
    spec_file.close()

    return columns

##TODO: hook this idea up -- change method
def _disambiguate_column_names(column_header, info_header):
    overlap = 0
    for column in info_header:
        if column in column_header:
            overlap = 1

    return ["INFO_" + i for i in info_header] if overlap else info_header

#TODO: cgates: suspect this could be simplified?
def _create_row_dict(column_list, vcf_record):
    row_dict = {"CHROM" : vcf_record.chrom,
                "POS" : vcf_record.pos,
                "ID" : vcf_record.vcf_id,
                "REF" : vcf_record.ref,
                "ALT" : vcf_record.alt,
                "QUAL" : vcf_record.qual,
                "FILTER" : vcf_record.filter}

    for sample_name in column_list[9:]:
        format_key_values = vcf_record.sample_tag_values[sample_name]
        for format_key, format_value in format_key_values.items():
            row_dict[format_key + "|" + sample_name] = format_value

    new_dict = row_dict.copy()
    new_dict.update(vcf_record.info_dict)

    return new_dict

def _filter_column_list(column_spec_list,
                        potential_col_list,
                        column_spec_filename):

    actual_column_list = []
    glossary_fields = OrderedDict()
    for i, column_regex in enumerate(column_spec_list):
        no_columns_found = True
        for (column_name, vcf_tag) in potential_col_list.items():
            column_exists = column_name in actual_column_list
            regex_matches = re.match("^" + column_regex + "$", column_name)
            if regex_matches and not column_exists:
                glossary_fields[vcf_tag] = 1
                actual_column_list.append(column_name)
                no_columns_found = False

        if no_columns_found:
            logger.warning(UNUSED_REGEX_WARNING_FORMAT,
                           column_regex,
                           column_spec_filename,
                           i + 1)

    if not actual_column_list:
        raise utils.JQException("The column specification file [{}] would "
                                "exclude all input columns. Review "
                                "inputs/usage and try again.",
                                column_spec_filename)
    glossary_fields = [x for x in glossary_fields.keys() if x]
    return actual_column_list, glossary_fields


def _create_potential_column_list(vcf_reader):
    fixed_fields = vcf_reader.split_column_header[0:8]
    columns = OrderedDict(zip(fixed_fields, [None] * 8))

    for info_tag in sorted(vcf_reader.info_metaheaders.keys()):
        columns[info_tag] = info_tag

    for format_tag in sorted(vcf_reader.format_metaheaders.keys()):
        for sample_name in vcf_reader.sample_names:
            columns[format_tag + "|" + sample_name] = format_tag

    return columns

def _get_glossary_writer(output_file):
    output_dir = os.path.dirname(output_file)
    complete_output_fname = os.path.basename(output_file)
    output_fname, dummy = os.path.splitext(complete_output_fname)
    glossary_fname = ".".join([output_fname, "glossary.txt"])
    glossary = os.path.join(output_dir, glossary_fname)

    return vcf.FileWriter(glossary)

def _create_glossary(metaheaders, glossary_fields, writer):
    glossary_entries = dict([_create_glossary_entry(x) for x in metaheaders])
    writer.write("FIELD_NAME\tTYPE\tDESCRIPTION\n")
    for field_name in glossary_fields:
        writer.write(glossary_entries[field_name])

def _create_glossary_entry(metaheader):
    try:
        header_type = re.search(r'^##(.*)=<', metaheader).group(1)
        id_value = re.search(r'^##.*\WID=(\w*)', metaheader).group(1)
        desc = re.search(r'^##.*Description="([^"]*)', metaheader).group(1)
        entry = "\t".join([id_value, header_type, desc.strip('"')]) + "\n"
        return (id_value, entry)
    except AttributeError:
        return (None, None)


def add_subparser(subparser):
    # pylint: disable=C0301
    parser = subparser.add_parser("expand", formatter_class=argparse.RawTextHelpFormatter, help="Pivots annotated VCF file so that given sample specific information is fielded out into separate columns. Returns an Excel file containing concatenation of all input files.")
    parser.add_argument("input", help="Path to annotated VCF file or path to directory of annotated VCF files. Other file types ignored")
    parser.add_argument("output", help="Path to directory of output variant-level TXT files")
    parser.add_argument("-v", "--verbose", action='store_true')
    parser.add_argument("-c", "--column_specification",
                        help="Path to text file containing column regular expressions to be included in output file",)
    parser.add_argument("--force", action='store_true', help="Overwrite contents of output directory")
    parser.add_argument("--log_file", help="Log file destination")

def _predict_output(args):
    return set([os.path.basename(args.output)])

def report_prediction(args):
    return _predict_output(args)

def get_required_input_output_types():
    return ("file", "file")

def validate_args(args):
    if args.column_specification:
        if not os.path.isfile(args.column_specification):
            raise utils.UsageError(("The column specification file [{}] could "
                                    "not be read. Review inputs/usage and "
                                    "try again."),
                                    args.column_specification)

        _read_col_spec(args.column_specification)

def _get_actual_columns(vcf_reader, col_spec):
    columns  = _create_potential_column_list(vcf_reader)
    glossary_fields = []
    for glossary_field in sorted([x for x in columns.values() if x], 
                                 key=lambda x: x.upper()):
        if glossary_field and glossary_field not in glossary_fields:
            glossary_fields.append(glossary_field)

    if col_spec:
        col_spec_columns = _read_col_spec(col_spec)
        (columns,
         glossary_fields) = _filter_column_list(col_spec_columns,
                                                columns,
                                                col_spec)
    return columns, glossary_fields


def execute(args, dummy_execution_context):
    #for the moment, there is no good place to put the execution context
    input_file = os.path.abspath(args.input)
    output_file = os.path.abspath(args.output)
    col_spec = args.column_specification if args.column_specification else None

    logger.debug("Expanding [{}] to [{}]",
                 input_file,
                 output_file)
    logger.info("Expanding [{}] to [{}]",
                args.input,
                args.original_output)

    vcf_reader = vcf.VcfReader(vcf.FileReader(input_file))
    file_writer = vcf.FileWriter(output_file)
    file_writer.open()

    (columns, glossary_fields) = _get_actual_columns(vcf_reader, col_spec)

    file_writer.write("#" + "\t".join(columns) + "\n")

    line_count = 0
    vcf_reader.open()
    for vcf_record in vcf_reader.vcf_records():
        row_dict = _create_row_dict(vcf_reader.split_column_header,
                                    vcf_record)

        new_line = []
        for col in columns:
            if col in row_dict:
                new_line.append(row_dict[col])
            else:
                new_line.append(".")

        file_writer.write("\t".join(new_line) + "\n")
        line_count +=1
        if line_count % 10000 == 0:
            logger.info("Expanding: {} rows processed", line_count)
    logger.info("Expand complete: {} rows processed", line_count)

    file_writer.close()

    glossary_writer = _get_glossary_writer(output_file)
    glossary_writer.open()
    _create_glossary(vcf_reader.metaheaders, glossary_fields, glossary_writer)
    glossary_writer.close()
    logger.info("Wrote glossary to [{}]",
                os.path.basename(glossary_writer.output_filepath))

    vcf_reader.close()
    logger.debug("Wrote input [{}] to output [{}]", input_file, output_file)



"""Validates command preconditions.

Specifically checks that the command, arguments, and environment
(e.g. input/output directories or files) are consistent and plausible.

Each validation function evaluates a specific precondition.
Each function is allowed to:
 * change the environment (e.g. create a dir)
 * change the args (replace the original output dir with a new temp output dir)
 * add arguments which may be required for sub-commands
 * delegate to/interact with a sub-command
 * raise a UsageException if things look problematic
"""
from __future__ import print_function, absolute_import, division

import errno
import glob
import os
import time

import jacquard.utils.utils as utils


_TEMP_WORKING_DIR_FORMAT = "jacquard.{}.{}.tmp"


def _actual_type(path):
    if os.path.isdir(path):
        return "directory"
    else:
        return "file"

def _build_collision_message(command, collisions):
    total_collisions = len(collisions)
    if total_collisions == 1:
        return ("The {} command would "
                "overwrite the existing file [{}]; review "
                "command/output dir to avoid overwriting or "
                "use the flag '--force'.").format(command,
                                                  collisions[0])
    cutoff = 5
    collision_list = ", ".join(collisions[0:min(cutoff, total_collisions)])
    if total_collisions > cutoff:
        omitted = total_collisions - cutoff
        collision_list += ", ...({} file(s) omitted)".format(omitted)
    return ("The {} command would "
            "overwrite {} existing files [{}]; review "
            "command/output dir to avoid overwriting or "
            "use the flag '--force'.").format(command,
                                              total_collisions,
                                              collision_list)

def _check_input_correct_type(dummy, args):
    module_name = args.subparser_name
    input_path = args.input
    required_type = args.required_input_type
    actual_type = _actual_type(input_path)
    if required_type != actual_type:
        raise utils.UsageError(("The {} command requires a {} as "
                                "input, but the specified input [{}] is a {}. "
                                "Review inputs and try again.") \
                                 .format(module_name,
                                         required_type,
                                         input_path,
                                         actual_type))

def _check_input_exists(dummy, args):
    if not os.path.exists(args.input):
        raise utils.UsageError(("Specified input [{}] does not exist. Review "\
                                 "inputs and try again.").format(args.input))

def _check_input_readable(dummy, args):
    try:
        if os.path.isdir(args.input):
            os.listdir(args.input)
        else:
            open(args.input, "r").close()
    except (OSError, IOError):
        raise utils.UsageError(("Specified input [{}] cannot be read. Review "
                                "inputs and try again.").format(args.input))

def _check_output_correct_type(module_name, output_path, required_type):
    actual_type = _actual_type(output_path)
    if required_type != actual_type:
        raise utils.UsageError(("The {} command outputs a {}, but the "
                                "specified output [{}] is a {}. "
                                "Review inputs and try again.")\
                                 .format(module_name,
                                         required_type,
                                         output_path,
                                         actual_type))

def _check_output_exists(dummy, args):
    if os.path.exists(args.output_path):
        _check_output_correct_type(args.subparser_name,
                                   args.output_path,
                                   args.required_output_type)

def _check_overwrite_existing_files(module, args):
    output = args.output_path
    if not os.path.isdir(output):
        output = os.path.dirname(output)

    existing_output_paths = sorted(glob.glob(os.path.join(output, "*")))
    existing_output = set([os.path.basename(i) for i in existing_output_paths])
    predicted_output = module.report_prediction(args)
    collisions = sorted(list(existing_output.intersection(predicted_output)))

    if collisions and not args.force:
        message = _build_collision_message(args.subparser_name, collisions)
        raise utils.UsageError(message)

def _check_there_will_be_output(module, args):
    predicted_output = module.report_prediction(args)

    if not predicted_output:
        message = ("Executing the {} command with the input [{}] would not "
                   "create any output files. Review inputs and try again.")\
                   .format(args.subparser_name, args.input)
        raise utils.UsageError(message)

def _check_valid_args(module, args):
    module.validate_args(args)

def _create_temp_working_dir(dummy, args):
    try:
        _makepath(args.temp_working_dir)
        if args.required_output_type == "directory":
            _makepath(args.output)
    except OSError:
        parent_dir = os.path.dirname(args.temp_working_dir)
        raise utils.UsageError(("Jacquard cannot write to output directory "
                                "[{}]. Review inputs and try again.")\
                                .format(parent_dir))

def _set_temp_working_dir(dummy, args):
    original_output = args.original_output
    required_output = args.required_output_type
    abs_original_output = os.path.abspath(original_output)
    pid = os.getpid()
    microseconds_since_epoch = int(time.time() * 1000 * 1000)
    dir_name = _TEMP_WORKING_DIR_FORMAT.format(str(pid),
                                               str(microseconds_since_epoch))

    place_temp_inside_output = True
    if required_output == "file":
        place_temp_inside_output = False
    elif required_output == "directory" and not os.path.isdir(original_output):
        place_temp_inside_output = False

    if place_temp_inside_output:
        base_dir = abs_original_output
        temp_working_dir = os.path.join(base_dir, dir_name)
        new_output = temp_working_dir
    else:
        base_dir = os.path.dirname(abs_original_output)
        temp_working_dir = os.path.join(base_dir, dir_name)
        new_output = os.path.join(temp_working_dir,
                                  os.path.basename(abs_original_output))

    args.temp_working_dir = temp_working_dir
    args.output = new_output

def _makepath(path):
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else: raise

def _set_required_types(module, args):
    (args.required_input_type,
     args.required_output_type) = module.get_required_input_output_types()

def _set_output_paths(dummy, args):
    args.original_output = args.output
    args.output_path = os.path.abspath(args.original_output)

_VALIDATION_TASKS = [_set_output_paths,
                     _set_required_types,
                     _set_temp_working_dir,
                     _check_input_exists,
                     _check_input_readable,
                     _check_input_correct_type,
                     _check_output_exists,
                     _create_temp_working_dir,
                     _check_there_will_be_output,
                     _check_overwrite_existing_files,
                     _check_valid_args]

def preflight(command, args):
    for validate in _VALIDATION_TASKS:
        validate(command, args)



"""Logs messages to console and file.

Error, warning, info messages are written to console (stderr) and file.
Debug messages are written to file unless logger is initialized as verbose (in
which case debug is also echoed to console).
"""
#pylint: disable=invalid-name, global-statement
from __future__ import print_function, absolute_import, division

from datetime import datetime
import errno
import getpass
import logging
import os
import socket
import sys

import jacquard.utils.utils as utils


WARNING_OCCURRED = False
"""Used to vary the Done message to emphasize upstream log warnings"""
log_filename = None


_DATE_FORMAT = '%Y-%m-%d %H:%M:%S'
_FILE_LOG_FORMAT = ('%(asctime)s|%(levelname)s|%(start_time)s|%(host)s|%(user)s'
                    '|%(tool)s|%(message)s')
_CONSOLE_LOG_FORMAT = '%(asctime)s|%(levelname)s|%(tool)s|%(message)s'
_logging_dict = {}
_verbose = False

def _makepath(path):
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else: raise

def _validate_log_file(log_file):
    try:
        _makepath(os.path.dirname(log_file))
    except OSError:
        raise utils.UsageError(("Jacquard cannot create specified log file "
                                "[{}]. Review inputs and try again."), log_file)

    try:
        log = open(log_file, "w")
        log.close()
    except IOError:
        raise utils.UsageError(("Jacquard cannot create specified log file "
                               "[{}]. Review inputs and try again."), log_file)

def initialize_logger(args):
    """Sets command name and formatting for subsequent calls to logger"""

    global log_filename
    log_filename = os.path.join(os.getcwd(), "jacquard.log")
    if args.log_file:
        _validate_log_file(args.log_file)
        log_filename = args.log_file

    logging.basicConfig(format=_FILE_LOG_FORMAT,
                        level="DEBUG",
                        datefmt=_DATE_FORMAT,
                        filename=log_filename)

    global _verbose
    if args.verbose:
        _verbose = args.verbose

    start_time = datetime.now().strftime(_DATE_FORMAT)
    global _logging_dict
    _logging_dict = {'user': getpass.getuser(),
                     'host': socket.gethostname(),
                     'start_time': start_time,
                     'tool': args.subparser_name}

def error(message, *args):
    _print("ERROR", message, args)
    logging.error(_format(message, args), extra=_logging_dict)

def warning(message, *args):
    _print("WARNING", message, args)
    logging.warning(_format(message, args), extra=_logging_dict)
    global WARNING_OCCURRED
    WARNING_OCCURRED = True

def info(message, *args):
    _print("INFO", message, args)
    logging.info(_format(message, args), extra=_logging_dict)

def debug(message, *args):
    if _verbose:
        _print("DEBUG", message, args)
    logging.debug(_format(message, args), extra=_logging_dict)

def _print(level, message, args):
    now = datetime.now().strftime(_DATE_FORMAT)
    print(_CONSOLE_LOG_FORMAT % {'asctime': now,
                                 'levelname':level,
                                 'tool':_logging_dict['tool'],
                                 'message': _format(message, args)},
          file=sys.stderr)
    sys.stderr.flush()

def _format(message, args):
    try:
        log_message = message.format(*[str(i) for i in args])
    except IndexError as err:
        log_message = ("Malformed log message ({}: {})"
                       "|{}|{}").format(type(err).__name__,
                                        err,
                                        message,
                                        [str(i) for i in args])
    return log_message



"""Classes to summarize zscore data for a sample-variant.

A collection of individual Tag classes hold the metaheader and logic to
transform Jacquard-standardized VcfRecords.

These transforms combine info from a single sample-variant tag with aggregate
info about all sample-variants in the VCF. For example, the zscore of a
sample-variant depth is:
    (my_depth - average of all depths) / stddev of all depths
For this reason, zscore values require the full VCF as one of the inputs.
"""
from __future__ import print_function, absolute_import, division

import math

from jacquard import __version__
import jacquard.utils.utils as utils


_JQ_SUMMARY_TAG = "JQ_SUMMARY_"

#TODO: (cgates): Define tag ids as public, class-level constants so dependent
# tags can reference them directly
class _AlleleFreqZScoreTag(object):
    TAG_ID = "{0}AF_ZSCORE".format(_JQ_SUMMARY_TAG)
    _RANGE_TAG = "{0}AF_RANGE".format(_JQ_SUMMARY_TAG)
    _METAHEADER_DESCRIPTION = (\
'''Concordance of reported allele frequencies across callers:
 [(this AF range - mean AF range)/standard dev(all AF ranges)].
 Values with null or missing AF range will be assigned zscore of \'.\';
 for multi-valued ranges, zscore is of largest range.''').replace("\n", "")

    def __init__(self, vcf_reader):
        self.tag_id = self.TAG_ID
        self.tag = _ZScoreTag(self.TAG_ID,
                              self._METAHEADER_DESCRIPTION,
                              self._RANGE_TAG,
                              vcf_reader)

    @property
    def metaheaders(self):
        return self.tag.metaheaders

    def add_tag_values(self, vcf_record):
        self.tag.add_tag_values(vcf_record)

class _DepthZScoreTag(object):
    TAG_ID = "{0}DP_ZSCORE".format(_JQ_SUMMARY_TAG)
    _RANGE_TAG = "{0}DP_RANGE".format(_JQ_SUMMARY_TAG)
    _METAHEADER_DESCRIPTION = ('''
Concordance of reported depth across callers: 
[(this DP range - mean DP range)/standard dev(all DP ranges)]. 
Values with null or missing DP range will be assigned zscore 
of \'.\'.''').replace("\n", "")

    def __init__(self, vcf_reader):
        self.tag_id = self.TAG_ID
        self.tag = _ZScoreTag(self.TAG_ID,
                              self._METAHEADER_DESCRIPTION,
                              self._RANGE_TAG,
                              vcf_reader)

    @property
    def metaheaders(self):
        return self.tag.metaheaders

    def add_tag_values(self, vcf_record):
        self.tag.add_tag_values(vcf_record)

class _ZScoreTag(object):
    '''Utility tag to add zscore for dependent tag (e.g. depth or allele freq)

    Given a dependent tag and a vcf reader, calculates mean and stdev on
    construction and then adds zscores for each value.
    '''
    _EXECUTION_FORMAT = "##jacquard.summarize.{0}.{1}_{2}={3}"
    _METAHEADER_FORMAT = ('##FORMAT=<ID={0},'
                          'Number=1,'
                          'Type=Float,'
                          'Description="{1}">')
    _MAX_PRECISION = 13

    def __init__(self,
                 tag_id,
                 metaheader_description,
                 dependent_tag_id,
                 vcf_reader):
        self._tag_id = tag_id
        self._dependent_tag_id = dependent_tag_id
        self._mean, self._stdev = self._init_population_stats(vcf_reader,
                                                              dependent_tag_id)
        self._metaheaders = self._init_metaheaders(tag_id,
                                                   metaheader_description,
                                                   dependent_tag_id,
                                                   self._mean,
                                                   self._stdev)

    def _init_metaheaders(self,
                          tag_id,
                          metaheader_description,
                          dependent_tag_id,
                          mean,
                          stdev):
        #pylint: disable=too-many-arguments
        metaheaders = []
        metaheaders.append(self._EXECUTION_FORMAT.format(tag_id,
                                                         dependent_tag_id,
                                                         "mean",
                                                         repr(mean)))
        metaheaders.append(self._EXECUTION_FORMAT.format(tag_id,
                                                         dependent_tag_id,
                                                         "stdev",
                                                         repr(stdev)))
        tag_metaheader = self._METAHEADER_FORMAT.format(tag_id,
                                                        metaheader_description,
                                                        __version__)
        metaheaders.append(tag_metaheader)
        return tuple(metaheaders)

    @property
    def metaheaders(self):
        return self._metaheaders


    def _ok_to_add_tag_values(self, vcf_record):
        return self._stdev and self._dependent_tag_id in vcf_record.format_tags


    @staticmethod
    def _zscore_as_str(zscore):
        if zscore == ".":
            return zscore
        else:
            return utils.round_two_digits(repr(zscore))

    def add_tag_values(self, vcf_record):
        if not self._ok_to_add_tag_values(vcf_record):
            return

        sample_values = {}
        for sample_name in vcf_record.sample_tag_values:
            zscore = "."
            tag_values = vcf_record.sample_tag_values[sample_name]
            value = self._get_dependent_value(tag_values,
                                              self._dependent_tag_id)

            if  value is not None:
                zscore = (value - self._mean) / self._stdev
            sample_values[sample_name] = self._zscore_as_str(zscore)

        vcf_record.add_sample_tag_value(self._tag_id,
                                        sample_values)

    @staticmethod
    def _get_dependent_value(tag_values, dependent_tag_id):
        '''Extract (float) value of dependent tag or None if absent.'''
        try:
            values = tag_values[dependent_tag_id].split(",")
            return max([float(value) for value in values])
        except KeyError:
            return None
        except ValueError:
            return None

    def _init_population_stats(self, vcf_reader, dependent_tag_id):
        '''Derive mean and stdev. 

        Adapted from online variance algorithm from Knuth, The Art of Computer 
        Programming, volume 2

        Returns: mean and stdev when len(values) > 1, otherwise (None, None)
            Values rounded to _MAX_PRECISION to ameliorate discrepancies between
            python versions.'''
        #pylint: disable=invalid-name
        n = 0
        mean = 0
        M2 = 0
        try:
            vcf_reader.open()
            for vcf_record in vcf_reader.vcf_records():
                for tag_values in vcf_record.sample_tag_values.values():
                    value = self._get_dependent_value(tag_values,
                                                      dependent_tag_id)
                    if value is not None:
                        n += 1
                        delta = value - mean
                        mean += delta / n
                        M2 += delta * (value - mean)
        finally:
            vcf_reader.close()

        mean = round(mean, self._MAX_PRECISION)

        stdev = 0
        if n == 0:
            mean = None
            stdev = None
        elif n >= 2:
            variance = M2/n
            stdev = round(math.sqrt(variance), self._MAX_PRECISION)

        return mean, stdev


class ZScoreCaller(object):
    """Provides metaheaders for VcfReader; adds summary tags to VcfRecords."""
    def __init__(self, vcf_reader):
        self._tags = [_AlleleFreqZScoreTag(vcf_reader),
                      _DepthZScoreTag(vcf_reader)]
        self._metaheaders = self._init_metaheaders(self._tags)

    @staticmethod
    def _init_metaheaders(tags):
        metaheaders = []
        for tag in tags:
            metaheaders.extend(tag.metaheaders)
        return tuple(metaheaders)

    @property
    def metaheaders(self):
        return self._metaheaders

    def add_tags(self, vcf_record):
        for tag in self._tags:
            tag.add_tag_values(vcf_record)
        return vcf_record.text()




"""Interprets MuTect VCF files adding Jacquard standard information.

MuTect VCFs are assumed to have a ".vcf" extension and have a valid
"##MuTect=..." metaheader.
"""
from __future__ import print_function, absolute_import, division
import jacquard.utils.utils as utils
import jacquard.variant_caller_transforms.common_tags as common_tags
import jacquard.utils.vcf as vcf

JQ_MUTECT_TAG = "JQ_MT_"
VERSION = "v1.1.4"

class _AlleleFreqTag(object):
    #pylint: disable=too-few-public-methods
    def __init__(self):
        self.metaheader = ('##FORMAT=<ID={0}AF,'
                           'Number=A,'
                           'Type=Float,'
                           #pylint: disable=line-too-long
                           'Description="Jacquard allele frequency for MuTect: Decimal allele frequency rounded to 2 digits (based on FA)">')\
                           .format(JQ_MUTECT_TAG)

    def add_tag_values(self, vcf_record):
        if "FA" in vcf_record.format_tags:
            sample_values = {}
            for sample in vcf_record.sample_tag_values:
                freq = vcf_record.sample_tag_values[sample]["FA"].split(",")
                sample_values[sample] = self._standardize_af(freq)
            vcf_record.add_sample_tag_value(JQ_MUTECT_TAG + "AF", sample_values)

    @staticmethod
    def _standardize_af(value):
        new_values = []
        for val in value:
            new_values.append(utils.round_two_digits(val))
        return ",".join(new_values)

class _DepthTag(object):
    #pylint: disable=too-few-public-methods
    def __init__(self):
        self.metaheader = ('##FORMAT=<ID={0}DP,'
                           'Number=1,'
                           'Type=Float,'
                           #pylint: disable=line-too-long
                           'Description="Jacquard depth for MuTect (based on DP)">')\
                           .format(JQ_MUTECT_TAG)

    @staticmethod
    def add_tag_values(vcf_record):
        if "DP" in vcf_record.format_tags:
            sample_values = {}
            for samp in vcf_record.sample_tag_values:
                sample_values[samp] = vcf_record.sample_tag_values[samp]["DP"]
            vcf_record.add_sample_tag_value(JQ_MUTECT_TAG + "DP", sample_values)

class _SomaticTag(object):
    #pylint: disable=too-few-public-methods
    def __init__(self):
        self.metaheader = ('##FORMAT=<ID={0}HC_SOM,'
                           'Number=1,'
                           'Type=Integer,'
                           #pylint: disable=line-too-long
                           'Description="Jacquard somatic status for MuTect: 0=non-somatic,1=somatic (based on SS FORMAT tag)">')\
                           .format(JQ_MUTECT_TAG)

    def add_tag_values(self, vcf_record):
        mutect_tag = JQ_MUTECT_TAG + "HC_SOM"
        sample_values = {}
        if "SS" in vcf_record.format_tags:
            for sample in vcf_record.sample_tag_values:
                somatic_status = vcf_record.sample_tag_values[sample]["SS"]
                sample_values[sample] = self._somatic_status(somatic_status)
        else:
            for sample in vcf_record.sample_tag_values:
                sample_values[sample] = "0"
        vcf_record.add_sample_tag_value(mutect_tag, sample_values)

    @staticmethod
    def _somatic_status(ss_value):
        if ss_value == "2":
            return "1"
        else:
            return "0"

class Mutect(object):
    """Recognize and transform MuTect VCFs to standard Jacquard format.

    MuTect VCFs are blessedly compliant and straightforward to translate, with
    the following exception. The incoming header has the sample name values
    (derived from the input alignments). To play well with other callers like
    Strelka and VarScan, the sample headers are replaced with Normal and Tumor.
    """
    _MUTECT_METAHEADER_PREFIX = "##MuTect"
    _NORMAL_SAMPLE_KEY = "normal_sample_name"
    _TUMOR_SAMPLE_KEY = "tumor_sample_name"

    def __init__(self):
        self.name = "MuTect"
        self.abbr = "MT"

    ##TODO (cgates): deprecate; remove
    @staticmethod
    def validate_input_file(meta_headers, dummy_column_header):
        for line in meta_headers:
            if line.startswith(Mutect._MUTECT_METAHEADER_PREFIX):
                return True
        return False

    @staticmethod
    def _is_mutect_vcf(file_reader):
        if file_reader.file_name.lower().endswith(".vcf"):
            vcf_reader = vcf.VcfReader(file_reader)
            for metaheader in vcf_reader.metaheaders:
                if metaheader.startswith(Mutect._MUTECT_METAHEADER_PREFIX):
                    return True
        return False

    def claim(self, file_readers):
        """Recognizes and claims MuTect VCFs form the set of all input VCFs.

        Each defined caller has a chance to evaluate and claim all the incoming
        files as something that it can process.

        Args:
            file_readers: the collection of currently unclaimed files

        Returns:
            A tuple of unclaimed readers and MuTectVcfReaders.
        """
        unclaimed_readers = []
        vcf_readers = []
        for file_reader in file_readers:
            if self._is_mutect_vcf(file_reader):
                vcf_reader = vcf.VcfReader(file_reader)
                vcf_readers.append(_MutectVcfReader(vcf_reader))
            else:
                unclaimed_readers.append(file_reader)
        return (unclaimed_readers, vcf_readers)

    @staticmethod
    def _build_mutect_dict(metaheaders):
        mutect_dict = {}
        for metaheader in metaheaders:
            if metaheader.startswith(Mutect._MUTECT_METAHEADER_PREFIX):
                split_line = metaheader.strip('"').split(" ")
                for item in split_line:
                    split_item = item.split("=")
                    try:
                        mutect_dict[split_item[0]] = split_item[1]
                    except IndexError:
                        pass

        return mutect_dict

    def _get_new_column_header(self, vcf_reader):
        """Returns a standardized column header.

        MuTect sample headers include the name of input alignment, which is
        nice, but doesn't match up with the sample names reported in Strelka
        or VarScan. To fix this, we replace with NORMAL and TUMOR using the
        MuTect metadata command line to replace them correctly."""
        mutect_dict = self._build_mutect_dict(vcf_reader.metaheaders)

        new_header_list = []
        required_keys = set([self._NORMAL_SAMPLE_KEY, self._TUMOR_SAMPLE_KEY])
        mutect_keys = set(mutect_dict.keys())

        if not required_keys.issubset(mutect_keys):
            raise utils.JQException("Unable to determine normal "
                                    "and tumor sample ordering "
                                    "based on MuTect metaheader.")

        for field_name in vcf_reader.column_header.split("\t"):
            if field_name == mutect_dict[self._NORMAL_SAMPLE_KEY]:
                field_name = "NORMAL"
            elif field_name == mutect_dict[self._TUMOR_SAMPLE_KEY]:
                field_name = "TUMOR"
            new_header_list.append(field_name)

        return "\t".join(new_header_list)


class _MutectVcfReader(object):
    """Adapter that presents a MuTect VCF as a VcfReader.

    This follows the VcfReader interface, delegating calls to the base
    VcfReader, adjusting metaheaders, column header, and individual
    variants as appropriate.

    See VcfReader for more info.
    """
    def __init__(self, vcf_reader):
        self._vcf_reader = vcf_reader
        self._caller = Mutect()
        self.tags = [common_tags.ReportedTag(JQ_MUTECT_TAG),
                     common_tags.PassedTag(JQ_MUTECT_TAG),
                     _AlleleFreqTag(),
                     _DepthTag(),
                     _SomaticTag()]

    def _get_new_metaheaders(self):
        return [tag.metaheader for tag in self.tags]

    @property
    def caller_name(self):
        return self._caller.name

    @property
    def file_name(self):
        return self._vcf_reader.file_name

    def open(self):
        return self._vcf_reader.open()

    def close(self):
        return self._vcf_reader.close()

    @staticmethod
    def expected_file_format():
        return [""]

    @property
    def metaheaders(self):
        new_metaheaders = list(self._vcf_reader.metaheaders)
        new_metaheaders.extend(self._get_new_metaheaders())
        caller_metaheader = "##jacquard.translate.caller={0}".\
                format(self._caller.name)
        new_metaheaders.append(caller_metaheader)

        return new_metaheaders

    @property
    def column_header(self):
        return self._caller._get_new_column_header(self._vcf_reader)

    def tagged_vcf_records(self):
        for vcf_record in self._vcf_reader.vcf_records():
            yield vcf_record

    def vcf_records(self):
        for vcf_record in self._vcf_reader.vcf_records():
            yield self._add_tags(vcf_record)

    def _add_tags(self, vcf_record):
        for tag in self.tags:
            tag.add_tag_values(vcf_record)
        return vcf_record




"""Interprets Strelka VCF files adding Jacquard standard information.

* Strelka VCFs are assumed to have a ".vcf" extension and have a
    "##source=strelka" metaheader.
* Strelka produces a separate file for SNVs and indels. Jacquard can process
    either or both.
* Jacquard standard tags are based on tier 2 data and use different source tags
    based on whether the file is indel or snp.

See tag definitions for more info.
"""
from __future__ import print_function, absolute_import, division

from collections import defaultdict, OrderedDict
import os
import re

import jacquard.utils.logger as logger
import jacquard.utils.utils as utils
import jacquard.variant_caller_transforms.common_tags as common_tags
import jacquard.utils.vcf as vcf


JQ_STRELKA_TAG = "JQ_SK_"
VERSION = "v2.0.15"

class _AlleleFreqTag(object):
    #pylint: disable=too-few-public-methods
    def __init__(self):
        #pylint: disable=line-too-long
        self.metaheader = ('##FORMAT=<ID={0}AF,'
                           'Number=A,'
                           'Type=Float,'
                           'Description="Jacquard allele frequency for Strelka: Decimal allele frequency rounded to 2 digits (based on alt_depth/total_depth. Uses (TIR tier 2)/DP2 if available, otherwise uses (ACGT tier2 depth) / DP2)">')\
                           .format(JQ_STRELKA_TAG)

    @staticmethod
    def _get_tier2_base_depth(sample_format_dict, alt_allele):
        numerator = float(sample_format_dict[alt_allele + "U"].split(",")[1])
        tags = ["AU", "CU", "TU", "GU"]
        depth = 0

        for tag in tags:
            depth += float(sample_format_dict[tag].split(",")[1])
        freq = numerator/depth if depth != 0 else 0.0

        return freq

    def _get_snp_allele_freq_per_sample(self, vcf_record, sample):
        afs = []
        split_alt = vcf_record.alt.split(",")

        for alt_allele in split_alt:
            sample_format_dict = vcf_record.sample_tag_values[sample]
            freq = self._get_tier2_base_depth(sample_format_dict, alt_allele)

            rounded_af = self._standardize_af(str(freq))
            capped_af = min(rounded_af, "1.00")
            afs.append(capped_af)

        return afs

    def _get_indelallelefreq_per_sample(self, vcf_record, sample):
        afs = []
        indel_depths = vcf_record.sample_tag_values[sample]["TIR"]
        numerator = float(indel_depths.split(",")[1])
        denominator = float(vcf_record.sample_tag_values[sample]["DP2"])
        freq = numerator/denominator if denominator != 0 else 0.0

        rounded_af = self._standardize_af(str(freq))
        capped_af = min(rounded_af, "1.00")
        afs.append(capped_af)

        return afs

    def add_tag_values(self, vcf_record):
        sample_values = {}
        if vcf_record.alt == ".":
            for sample in vcf_record.sample_tag_values:
                sample_values[sample] = "."
        else:
            for sample in vcf_record.sample_tag_values:
                if "AU" in vcf_record.format_tags:#if it's an snp
                    afs = self._get_snp_allele_freq_per_sample(vcf_record,
                                                               sample)
                    sample_values[sample] = ",".join(afs)
                elif "TIR" in vcf_record.format_tags: #if it's an indel
                    afs = self._get_indelallelefreq_per_sample(vcf_record,
                                                               sample)
                    sample_values[sample] = ",".join(afs)
                else:
                    continue

        if sample_values:
            vcf_record.add_sample_tag_value(JQ_STRELKA_TAG + "AF",
                                            sample_values)

    @staticmethod
    def _standardize_af(value):
        return utils.round_two_digits(value)

class _DepthTag(object):
    #pylint: disable=too-few-public-methods
    REQUIRED_TAGS = set(["DP2", "AU"])
    NUCLEOTIDE_DEPTH_TAGS = ["AU", "CU", "TU", "GU"]

    @classmethod
    def _get_tier2_base_depth(cls, sample_tags):
        if "DP2" in sample_tags:
            return sample_tags["DP2"]
        depth = 0
        for tag in _DepthTag.NUCLEOTIDE_DEPTH_TAGS:
            tier2_depth = sample_tags[tag].split(",")[1]
            depth += int(tier2_depth)
        return str(depth)

    def __init__(self):
        #pylint: disable=line-too-long
        self.metaheader = ('##FORMAT=<ID={0}DP,'
                           'Number=1,'
                           'Type=Float,'
                           'Description="Jacquard depth for Strelka (uses DP2 if available, otherwise uses ACGT tier2 depth)">')\
                           .format(JQ_STRELKA_TAG)

    @staticmethod
    def add_tag_values(vcf_record):
        if vcf_record.format_tags.isdisjoint(_DepthTag.REQUIRED_TAGS):
            return
        sample_values = {}
        for sample in vcf_record.sample_tag_values:
            sample_tags = vcf_record.sample_tag_values[sample]
            sample_values[sample] = _DepthTag._get_tier2_base_depth(sample_tags)
        vcf_record.add_sample_tag_value(JQ_STRELKA_TAG + "DP", sample_values)


##TODO (cgates): Make this robust to sample order changes
class _SomaticTag(object):
    #pylint: disable=too-few-public-methods
    def __init__(self):
        #pylint: disable=line-too-long
        self.metaheader = ('##FORMAT=<ID={0}HC_SOM,'
                           'Number=1,Type=Integer,'
                           'Description="Jacquard somatic status for Strelka: 0=non-somatic,1=somatic (based on PASS in FILTER column)">')\
                           .format(JQ_STRELKA_TAG)

    @staticmethod
    def add_tag_values(vcf_record):
        sample_values = {}
        for i, sample in enumerate(vcf_record.sample_tag_values):
            sample_values[sample] = _SomaticTag._somatic_status(i, vcf_record)
        strelka_tag = JQ_STRELKA_TAG + "HC_SOM"
        vcf_record.add_sample_tag_value(strelka_tag, sample_values)

    @classmethod
    def _somatic_status(cls, sample_index, vcf_record):
        if sample_index == 1 and vcf_record.filter == "PASS":
            return "1"
        else:
            return "0"

class Strelka(object):
    """Recognize and transform Strelka VCFs to standard Jacquard format.

    Note that Strelka sometimes reports variants which fail the filter as
    having an ALT of "."; this seems to be Strelka's shorthand for saying,
    "I couldn't be sure there was an ALT". Unfortunately, that's not a valid
    ALT value so these rows are flagged and (in a typical workflow) excluded.
    """

    def __init__(self):
        self.name = "Strelka"
        self.abbr = "SK"
        self.meta_header = "##jacquard.normalize_strelka.sources={0},{1}\n"

    ##TODO (cgates): deprecated; remove
    @staticmethod
    def validate_input_file(meta_headers, dummy_column_header):
        return "##source=strelka" in meta_headers

    @staticmethod
    def _is_strelka_vcf(file_reader):
        if file_reader.file_name.endswith(".vcf"):
            vcf_reader = vcf.VcfReader(file_reader)
            return "##source=strelka" in vcf_reader.metaheaders
        return False

    def _find_strelka_files(self, file_readers):
        prefix_dict = OrderedDict()
        unclaimed_readers = []
        for file_reader in file_readers:
            if self._is_strelka_vcf(file_reader):
                prefix, _ = os.path.splitext(file_reader.file_name)
                prefix_dict[prefix] = file_reader
            else:
                unclaimed_readers.append(file_reader)

        return prefix_dict, unclaimed_readers

    @staticmethod
    def _split_prefix_by_patient(prefix_dict):
        prefix_by_patients = defaultdict(list)
        snv_indel = ["snvs", "indels"]

        for prefix in prefix_dict:
            patient_names = [i for i in prefix.split(".") if i not in snv_indel]
            patient_name = ".".join(patient_names)
            prefix_by_patients[patient_name].append(prefix)

        return prefix_by_patients

    @staticmethod
    def _validate_vcf_readers(prefix_by_patients):
        number_of_files = set()
        for file_names in prefix_by_patients.values():
            if len(file_names) == 1:
                for file_name in file_names:
                    if re.search("snvs", file_name):
                        msg = "Strelka VCF [{}] has no indels file."
                        logger.error(msg, file_name)
                    elif re.search("indels", file_name):
                        msg = "Strelka VCF [{}] has no snvs file."
                        logger.error(msg, file_name)
            number_of_files.add(len(file_names))

        if len(number_of_files) > 1:
            msg = ("Some Strelka VCFs were missing either a snvs or indels "
                   "file. Review inputs/command options and try again.")
            raise utils.JQException(msg)

    @staticmethod
    def _create_vcf_readers(prefix_to_readers):
        vcf_readers = []
        for file_reader in prefix_to_readers.values():
            vcf_reader = vcf.VcfReader(file_reader)
            vcf_readers.append(_StrelkaVcfReader(vcf_reader))

        return vcf_readers

    def claim(self, file_readers):
        """Recognizes and claims Strelka VCFs form the set of all input VCFs.

        Each defined caller has a chance to evaluate and claim all the incoming
        files as something that it can process.

        Args:
            file_readers: the collection of currently unclaimed files

        Returns:
            A tuple of unclaimed readers and StrelkaVcfReaders.
        """
        (prefix_to_reader,
         unclaimed_readers) = self._find_strelka_files(file_readers)
        prefix_by_patients = self._split_prefix_by_patient(prefix_to_reader)
        self._validate_vcf_readers(prefix_by_patients)
        vcf_readers = self._create_vcf_readers(prefix_to_reader)

        return (unclaimed_readers, vcf_readers)

class _StrelkaVcfReader(object):
    """Adapter that presents a Strelka VCF as a VcfReader.

    This follows the VcfReader interface, delegating calls to the base
    VcfReader, adjusting metaheaders, and individual
    variants as appropriate.

    See VcfReader for more info.
    """
    def __init__(self, vcf_reader):
        self._vcf_reader = vcf_reader
        self._caller = Strelka()
        self.tags = [common_tags.ReportedTag(JQ_STRELKA_TAG),
                     common_tags.PassedTag(JQ_STRELKA_TAG),
                     _AlleleFreqTag(),
                     _DepthTag(),
                     _SomaticTag()]

    def _get_new_metaheaders(self):
        return [tag.metaheader for tag in self.tags]

    @property
    def caller_name(self):
        return self._caller.name

    @property
    def file_name(self):
        return self._vcf_reader.file_name

    def open(self):
        return self._vcf_reader.open()

    def close(self):
        return self._vcf_reader.close()

    @staticmethod
    def expected_file_format():
        return ["snvs", "indels"]

    @property
    def metaheaders(self):
        new_metaheaders = list(self._vcf_reader.metaheaders)
        new_metaheaders.extend(self._get_new_metaheaders())
        caller_metaheader = "##jacquard.translate.caller={0}".\
                format(self._caller.name)
        new_metaheaders.append(caller_metaheader)

        return new_metaheaders

    @property
    def column_header(self):
        return self._vcf_reader.column_header

    def tagged_vcf_records(self):
        for vcf_record in self._vcf_reader.vcf_records():
            yield vcf_record

    def vcf_records(self):
        for vcf_record in self._vcf_reader.vcf_records():
            yield self._add_tags(vcf_record)

    def _add_tags(self, vcf_record):
        for tag in self.tags:
            tag.add_tag_values(vcf_record)
        return vcf_record




"""Classes to parse, interpret, and manipulate VCF files and records."""
from __future__ import print_function, absolute_import, division

from collections import OrderedDict
import os
import re
import sys

import natsort

import jacquard.utils.utils as utils


#TODO: (cgates): add context management to open/close
class VcfReader(object):
    """Read only wrapper providing VCF metaheaders and records."""
    def __init__(self, file_reader):
        self._file_reader = file_reader
        (self.column_header, self.metaheaders) = self._init_headers()
        self.split_column_header = self.column_header.strip("#").split("\t")
        self.sample_names = self._init_sample_names()
        self.qualified_sample_names = self._create_qualified_sample_names()

    def __lt__(self, other):
        return self._file_reader < other._file_reader

    def _get_tag_metaheaders(self, regex_exp):
        tag_dict = {}
        for metaheader in self.metaheaders:
            tag = re.match(regex_exp, metaheader)
            if tag:
                tag_key = tag.group(1)
                tag_dict[tag_key] = metaheader.strip()

        return tag_dict

    @property
    def file_name(self):
        return self._file_reader.file_name

    @property
    def input_filepath(self):
        return self._file_reader.input_filepath

    @property
    def format_metaheaders(self):
        return dict(self._get_tag_metaheaders("^##FORMAT=.*?[<,]ID=([^,>]*)"))

    @property
    def info_metaheaders(self):
        return dict(self._get_tag_metaheaders("^##INFO=.*?[<,]ID=([^,>]*)"))

    @property
    def filter_metaheaders(self):
        return dict(self._get_tag_metaheaders("^##FILTER=.*?[<,]ID=([^,>]*)"))

    @property
    def contig_metaheaders(self):
        return dict(self._get_tag_metaheaders("^##contig=.*?[<,]ID=([^,>]*)"))

    def _init_sample_names(self):
        sample_names = []
        column_fields = self.column_header.split("\t")
        if len(column_fields) > 8:
            sample_names = column_fields[9:]

        return sample_names

    def _create_qualified_sample_names(self):
        patient_prefix = self.file_name.split(".")[0]
        qualified_names = []
        for sample_name in self.sample_names:
            qualified_names.append("|".join([patient_prefix, sample_name]))
        return qualified_names

    def _init_headers(self):
        metaheaders = []
        column_header = None
        try:
            self._file_reader.open()
            for line in self._file_reader.read_lines():
                if line.startswith("##"):
                    metaheaders.append(line.rstrip())
                elif line.startswith("#"):
                    column_header = line.rstrip()
                else:
                    break
        finally:
            self._file_reader.close()

        if not (column_header and metaheaders):
            raise utils.JQException("ERROR: [{}] is not a valid vcf. Missing "
                                    "column header or metaheaders."\
                                    .format(self.file_name))

        return column_header, tuple(metaheaders)

    #TODO (cgates): qualified is used by ONE invocation in merge. Can we
    #somehow make merge do this instead of universally complicating the method?
    def vcf_records(self, qualified=False):
        """Generates parsed VcfRecord objects.

        Typically called in a for loop to process each vcf record in a
        VcfReader. VcfReader must be opened in advanced and closed when
        complete. Skips all headers.

        Args:
            qualified: When True, sample names are prefixed with file name

        Returns:
            Parsed VcfRecord

        Raises:
            StopIteration: when reader is exhausted.
            TypeError: if reader is closed.
        """
        if qualified:
            sample_names = self.qualified_sample_names
        else:
            sample_names = self.sample_names

        for line in self._file_reader.read_lines():
            if line.startswith("#"):
                continue
            yield VcfRecord.parse_record(line, sample_names)

    def open(self):
        self._file_reader.open()

    def close(self):
        self._file_reader.close()

class VcfRecord(object): #pylint: disable=too-many-instance-attributes
    """Represents an specific variant record.

    All functionality for parsing, interpreting, or manipulating a VCF record
    should be contained in this class. Mutator methods are implemented to add
    fields/tags to existing record, but replace/remove fields/tags.

    Attributes and args typically mirror the VCF fields; a few are renamed to
    avoid collisions with Python-reserved words. Except where noted, all
    attributes are stored as strings.

    Attributes:
            chrom:
            pos:
            ref:
            alt:
            vcf_id:
            qual:
            vcf_filter:
            info_dict: dict of info fields key-value pairs (flags stored as
                key-None)
            sample_tag_values: dict of samples where each value is a dict of
                tag-values. All keys and values are stored as strings. Example:
                {sampleA : {AF:0.3, DP:42}, sampleB : {AF:0.1, DP:25}}
                Note that dict is a natural pythonic representation of sample
                tag values, it's sometimes helpful to think of sample_tag_values
                as a table of samples (rows) x format_tags (columns).
            _key: tuple that defines record equality and ordering.
    """
    _EMPTY_SET = set()
    _FILTERS_TO_REPLACE = set(["", ".", "pass"])

    @classmethod
    def parse_record(cls, vcf_line, sample_names):
        """Alternative constructor that parses VcfRecord from VCF string.

        Aspire to parse/represent the data such that it could be reliably
        round-tripped. (This nicety means INFO fields and FORMAT tags should be
        treated as ordered to avoid shuffling.)

        Args:
            vcf_line: the VCF variant record as a string; tab separated fields,
                trailing newlines are ignored. Must have at least 8 fixed fields
                (through INFO)
            sample_names: a list of sample name strings; these should match
                the VCF header column
        Returns:
            A mutable VcfRecord.
        """
        vcf_fields = vcf_line.rstrip("\n").split("\t")
        chrom, pos, rid, ref, alt, qual, rfilter, info \
                = vcf_fields[0:8]
        sample_fields = []
        sample_tag_values = {}
        if len(vcf_fields) > 9:
            rformat = vcf_fields[8]
            sample_fields = vcf_fields[9:]
            sample_tag_values = VcfRecord._sample_tag_values(sample_names,
                                                             rformat,
                                                             sample_fields)
        return VcfRecord(chrom, pos, ref, alt,
                         rid, qual, rfilter, info,
                         sample_tag_values)

    @classmethod
    def _sample_tag_values(cls, sample_names, rformat, sample_fields):
        """Creates a sample dict of tag-value dicts for a single variant record.

        Args:
            sample_names: list of sample name strings.
            rformat: record format string (from VCF record).
            sample_fields: list of strings where each string is the ';'
                seperated format values for an individual sample.

        Returns:
            An dict of samples, where each key is a sample and each value
                is an dict of format-values. See attribute below for example.
                Will return '.' if no values for sampe field.
        """
        sample_tag_values = OrderedDict()
        tag_names = VcfRecord._format_list(rformat)
        for i, sample_field in enumerate(sample_fields):
            tag_values = sample_field.split(":") if sample_field else "."
            sample_tag_values[sample_names[i]] = OrderedDict(zip(tag_names,
                                                                 tag_values))
        return sample_tag_values

    @classmethod
    def _format_list(cls, rformat):
        if rformat and rformat != ".":
            return rformat.split(":")
        else:
            return []

    @classmethod
    def _str_as_int(cls, string):
        if "chr" in string:
            string = string.replace("chr", "")
        try:
            return int(string)
        except ValueError:
            return sys.maxsize
#TODO: (cgates): Could we make filter an OrderedSet
#TODO: (cgates) adjust info field to be stored as dict only instead of string
#pylint: disable=too-many-arguments
    def __init__(self, chrom, pos, ref, alt,
                 vcf_id=".", qual=".", vcf_filter=".", info=".",
                 sample_tag_values=None):
        """Builds a mutable VcfRecord from constituent VCF fields.

        If building from a VCF record line, see VcfRecord.parse_record.
        """
        self.chrom = chrom
        self.pos = pos
        self.vcf_id = vcf_id
        self.ref = ref
        self.alt = alt
        self.qual = qual
        self.filter = vcf_filter
        self.info = info
        self.info_dict = self._init_info_dict()

        if sample_tag_values is None:
            self.sample_tag_values = OrderedDict()
        else:
            self.sample_tag_values = sample_tag_values
        self._key = self._build_key()

    def _build_key(self):
        return (VcfRecord._str_as_int(self.chrom),
                self.chrom,
                VcfRecord._str_as_int(self.pos),
                self.ref,
                self.alt)

    @property
    def format_tags(self):
        """Returns set of format tags."""
        tags = VcfRecord._EMPTY_SET
        if self.sample_tag_values:
            first_sample = list(self.sample_tag_values.keys())[0]
            tags = set(self.sample_tag_values[first_sample].keys())
        return tags

    def _init_info_dict(self):
        info_dict = OrderedDict()
        if self.info and self.info != ".":
            info_list = self.info.split(";")
            for key_value in info_list:
                if "=" in key_value:
                    key, value = key_value.split("=")
                    info_dict[key] = value
                else:
                    info_dict[key_value] = key_value
        return info_dict

    def add_info_field(self, field):
        """Adds new info field (flag or key=value pair).

        Args:
            field: String flag (e.g. "SOMATIC") or key-value ("NEW_DP=42")

        Raises:
            KeyError: if info field already exists
        """
        if field in self.info_dict:
            msg = "New info field [{}] already exists.".format(field)
            raise KeyError(msg)

        if "=" in field:
            key, value = field.split("=")
            self.info_dict[key] = value
        else:
            self.info_dict[field] = field

        self._join_info_fields()

    #TODO:(cgates): Remove info; all external calls should reference info_dict
    def _join_info_fields(self):
        """Updates info attribute from info dict."""
        if self.info_dict:
            info_fields = []
            if len(self.info_dict) > 1:
                self.info_dict.pop(".", None)
            for field, value in self.info_dict.items():
                if field == value:
                    info_fields.append(value)
                else:
                    info_fields.append("=".join([field, value]))
            self.info = ";".join(info_fields)
        else:
            self.info = "."

    #TODO cgates: move this to merge
    def get_empty_record(self):
        return VcfRecord(chrom=self.chrom,
                         pos=self.pos,
                         ref=self.ref,
                         alt=self.alt)

    def _format_field(self):
        """Returns string representation of format field."""
        format_field = "."
        if self.sample_tag_values:
            first_sample = list(self.sample_tag_values.keys())[0]
            tag_names = self.sample_tag_values[first_sample].keys()
            if tag_names:
                format_field = ":".join(tag_names)
        return format_field

    def _sample_field(self, sample):
        """Returns string representation of sample-format values.

        Raises:
            KeyError: if requested sample is not defined.
        """
        tag_values = self.sample_tag_values[sample].values()
        if tag_values:
            return ":".join(tag_values)
        else:
            return "."

    def text(self):
        "Returns tab-delimited, newline terminated string of VcfRecord."
        stringifier = [self.chrom, self.pos, self.vcf_id, self.ref, self.alt,
                       self.qual, self.filter, self.info,
                       self._format_field()]

        for sample in self.sample_tag_values:
            stringifier.append(self._sample_field(sample))

        return "\t".join(stringifier) + "\n"

    def _samples_match(self, new_sample_values):
        return set(new_sample_values.keys()) == \
                set(self.sample_tag_values.keys())

    def add_sample_tag_value(self, tag_name, new_sample_values):
        """Appends a new format tag-value for all samples.

        Args:
            tag_name: string tag name; must not already exist
            new_sample

        Raises:
            KeyError: if tag_name to be added already exists
        """
        if tag_name in self.format_tags:
            msg = "New format value [{}] already exists.".format(tag_name)
            raise KeyError(msg)

        if not self._samples_match(new_sample_values):
            raise KeyError("Sample name values must match "
                           "existing sample names")
        for sample in self.sample_tag_values.keys():
            value = str(new_sample_values[sample])
            self.sample_tag_values[sample][tag_name] = value

    def add_or_replace_filter(self, new_filter):
        """Replaces null or blank filter or adds filter to existing list."""
        if self.filter.lower() in self._FILTERS_TO_REPLACE:
            self.filter = new_filter
        elif new_filter not in self.filter.split(";"):
            self.filter = ";".join([self.filter,
                                    new_filter])

    def __eq__(self, other):
        return isinstance(other, VcfRecord) and self._key == other._key

    def __hash__(self):
        return hash(self._key)

    def __lt__(self, other):
        return self._key < other._key


#TODO cgates: add context management to open/close
class FileWriter(object):
    """Trivial wrapper around os file to expedite testing."""
    def __init__(self, output_filepath):
        self.output_filepath = output_filepath
        self._file_writer = None

    @property
    def file_name(self):
        return os.path.basename(self.output_filepath)

    def open(self):
        self._file_writer = open(self.output_filepath, "w")

    def write(self, text):
        return self._file_writer.write(text)

    def close(self):
        self._file_writer.close()

    def __eq__(self, other):
        return (isinstance(other, self.__class__)
                and self.__dict__ == other.__dict__)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.output_filepath)

class FileReader(object):
    """Trivial wrapper around os file to expedite testing/natural sorting."""

    def __init__(self, input_filepath):
        self.input_filepath = input_filepath
        self.file_name = os.path.basename(input_filepath)
        self._file_reader = None

    def open(self):
        self._file_reader = open(self.input_filepath, "r")

    def read_lines(self):
        for line in self._file_reader:
            yield line

    def close(self):
        self._file_reader.close()

    def __eq__(self, other):
        return (isinstance(other, self.__class__)
                and self.__dict__ == other.__dict__)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.input_filepath)

    def __lt__(self, other):
        key = natsort.natsort_keygen()
        return key(self.file_name) < key(other.file_name)




"""Common tags used by several callers."""
from __future__ import print_function, absolute_import, division


CALLER_REPORTED_TAG = "CALLER_REPORTED"
CALLER_PASSED_TAG = "CALLER_PASSED"

class ReportedTag(object):
    """Tracks whether the caller reported this variant (i.e. it's in the VCF).

    This tag could be inferred through the presence of other tags, but adding
    it explicitly simplifies how summary tags are generated.
    """
    #pylint: disable=too-few-public-methods
    def __init__(self, tag_name):
        self.tag_name = tag_name
        self.metaheader = ('##FORMAT=<ID={}{},'
                           'Number=1,'
                           'Type=Integer,'
                           #pylint: disable=line-too-long
                           'Description="1 = variant present in original VCF">')\
                           .format(self.tag_name,
                                   CALLER_REPORTED_TAG)

    def add_tag_values(self, vcf_record):
        sample_values = {}
        for sample in vcf_record.sample_tag_values:
            sample_values[sample] = 1
        vcf_record.add_sample_tag_value(self.tag_name+CALLER_REPORTED_TAG,
                                        sample_values)

class PassedTag(object):
    """Tracks whether the caller passed this variant.

    This enables a useful summary tag. Note that Jaquard may flag a variant
    that originally passed as invalid (e.g. invalid Varscan ALT value). In
    this case, this tag represents what the origin caller thought and not
    Jacquard's opinion.
    """
    #pylint: disable=too-few-public-methods
    def __init__(self, tag_name):
        self.tag_name = tag_name
        self.metaheader = ('##FORMAT=<ID={}{},'
                           'Number=1,Type=Integer,'
                           'Description="1 = variant FILTER is PASS in '
                           'original VCF">').format(self.tag_name,
                                                    CALLER_PASSED_TAG)

    def add_tag_values(self, vcf_record):
        sample_values = {}
        for sample in vcf_record.sample_tag_values:
            if vcf_record.filter == "PASS":
                sample_values[sample] = 1
            else:
                sample_values[sample] = 0
        vcf_record.add_sample_tag_value(self.tag_name + CALLER_PASSED_TAG,
                                        sample_values)



"""Interprets VarScan2 VCF files adding Jacquard standard information.

* VarScan VCFs are assumed to have a ".vcf" extension and have a
    "##source=VarScan2" metaheader.
* VarScan produces a separate file for SNPs and indels. Jacquard can process
    either or both.
* Jacquard requires the VarScan VCF outputs. The VarScan workflow has optional
    extra steps to:
    a) partition the results into Germline, LOH, and Somatic files
    b) filter to a subset of high-confidence variants
    If these files are provided in the input, Jacquard will use them to
    flag variants not found in the high-confidence files so that they can be
    optionally filtered out.
* If provided, high-confidence files should follow the naming convention of
    patientName.*.fpfilter.pass
    The high-confidence file suffix can be supplied as a command line arg.

See tag definitions for more info.
"""
from __future__ import print_function, absolute_import, division

from collections import defaultdict, OrderedDict
import os
import re

import jacquard.utils.logger as logger
import jacquard.utils.utils as utils
import jacquard.variant_caller_transforms.common_tags as common_tags
import jacquard.utils.vcf as vcf


_VARSCAN_SOMATIC_HEADER = ("#CHROM|POS|ID|REF|ALT|QUAL|FILTER|INFO|FORMAT|"
                           "NORMAL|TUMOR").replace("|", "\t")
JQ_VARSCAN_TAG = "JQ_VS_"
VERSION = "v2.3"

class _AlleleFreqTag(object):
    #pylint: disable=too-few-public-methods
    @classmethod
    def _standardize_af(cls, value):
        new_values = []
        for val in value:
            new_val = str(float(val.strip("%")) / 100)
            new_values.append(utils.round_two_digits(new_val))
        return ",".join(new_values)

    def __init__(self):
        #pylint: disable=line-too-long
        self.metaheader = ('##FORMAT=<ID={0}AF,'
                           'Number=A,'
                           'Type=Float,'
                           'Description="Jacquard allele frequency for VarScan: Decimal allele frequency rounded to 2 digits (based on FREQ)">')\
                           .format(JQ_VARSCAN_TAG)

    @staticmethod
    def add_tag_values(vcf_record):
        sample_values = {}
        if "FREQ" in vcf_record.format_tags:
            for sample in vcf_record.sample_tag_values:
                freq = vcf_record.sample_tag_values[sample]["FREQ"].split(",")
                sample_values[sample] = _AlleleFreqTag._standardize_af(freq)
            vcf_record.add_sample_tag_value(JQ_VARSCAN_TAG + "AF",
                                            sample_values)

class _DepthTag(object):
    #pylint: disable=too-few-public-methods
    def __init__(self):
        #pylint: disable=line-too-long
        self.metaheader = ('##FORMAT=<ID={0}DP,'
                           'Number=1,'
                           'Type=Float,'
                           'Description="Jacquard depth for VarScan (based on DP)">')\
                           .format(JQ_VARSCAN_TAG)

    @staticmethod
    def add_tag_values(vcf_record):
        if "DP" in vcf_record.format_tags:
            sample_values = {}
            for sample in vcf_record.sample_tag_values:
                depth = vcf_record.sample_tag_values[sample]["DP"]
                sample_values[sample] = depth
            vcf_record.add_sample_tag_value(JQ_VARSCAN_TAG + "DP",
                                            sample_values)

##TODO (cgates): Make this robust to sample order changes
class _SomaticTag(object):
    #pylint: disable=too-few-public-methods

    @staticmethod
    def _somatic_status(sample_index):
        if sample_index == 0:  # it's NORMAL
            return "0"
        else:  # it's TUMOR
            return "1"

    def __init__(self):
        #pylint: disable=line-too-long
        self.metaheader = ('##FORMAT=<ID={0}HC_SOM,'
                           'Number=1,'
                           'Type=Integer,'
                           'Description="Jacquard somatic status for VarScan: 0=non-somatic,1=somatic (based on SOMATIC info tag and if sample is TUMOR)">')\
                           .format(JQ_VARSCAN_TAG)

    @staticmethod
    def add_tag_values(vcf_record):
        info_array = vcf_record.info.split(";")
        varscan_tag = JQ_VARSCAN_TAG + "HC_SOM"
        sample_values = {}

        if "SS=2" in info_array and vcf_record.filter == "PASS":
            for i, sample in enumerate(vcf_record.sample_tag_values):
                sample_values[sample] = _SomaticTag._somatic_status(i)
        else:
            for sample in vcf_record.sample_tag_values:
                sample_values[sample] = "0"

        vcf_record.add_sample_tag_value(varscan_tag, sample_values)

#TODO: (cgates/jebene): All tags should have _TAG_ID as implemented below
class _HCTag(object):
    #pylint: disable=too-few-public-methods
    _FILTERS_TO_REPLACE = set(["", ".", "pass"])
    _TAG_ID = "{}LOW_CONFIDENCE".format(JQ_VARSCAN_TAG)

    def __init__(self, file_reader):
        #pylint: disable=line-too-long
        self.metaheader = ('##FILTER=<ID={},'
                           'Number=1,'
                           'Type=Flag,'
                           'Description="Jacquard high-confidence somatic flag for VarScan. Based on intersection with filtered VarScan variants">')\
                           .format(self._TAG_ID)
        self._hc_loci = self._parse_file_reader(file_reader)

    @staticmethod
    def _parse_file_reader(file_reader):
        column_header = None
        hc_loci = set()
        file_reader.open()
        for line in file_reader.read_lines():
            if line.startswith("chrom\tposition"):
                column_header = line
            else:
                split_line = line.split("\t")
                hc_loci.add((split_line[0], split_line[1]))
        file_reader.close()

        #TODO : (cgates): Please test this
        if not column_header:
            raise utils.JQException("Error. The hc file {} is in an incorrect"
                                    "format. Review inputs and try"
                                    "again.".format(file_reader.file_name))

        return hc_loci

    def add_tag_values(self, vcf_record):
        if (vcf_record.chrom, vcf_record.pos) not in self._hc_loci:
            vcf_record.add_or_replace_filter(self._TAG_ID)
        return vcf_record

class Varscan(object):
    """Recognize and transform VarScan VCFs to standard Jacquard format."""

    _DEFAULT_REGEX = "Somatic.hc.fpfilter.pass"

    def __init__(self, args=None):
        self.name = "VarScan"
        self.abbr = "VS"
        self.meta_header = "##jacquard.normalize_varscan.sources={0},{1}\n"

        self.hc_file_pattern = self._get_hc_file_pattern(args)

    @staticmethod
    def _get_hc_file_pattern(args):
        msg = ("The specified regex [{}] could not be compiled. "
               "Review inputs and try again")
        if args and args.varscan_hc_filter_file_regex:
            try:
                compiled_regex = re.compile(args.varscan_hc_filter_file_regex)
            except:
                raise utils.UsageError(msg, args.varscan_hc_filter_file_regex)
        else:
            try:
                compiled_regex = re.compile(Varscan._DEFAULT_REGEX)
            except:
                raise utils.UsageError(msg, Varscan._DEFAULT_REGEX)

        return compiled_regex

    ##TODO (cgates): deprecated; remove
    @staticmethod
    def validate_input_file(meta_headers, column_header):
        if "##source=VarScan2" not in meta_headers:
            return 0

        if _VARSCAN_SOMATIC_HEADER == column_header:
            return 1
        else:
            raise utils.JQException("Unexpected VarScan VCF structure - "
                                    "missing NORMAL and TUMOR headers.")

    @staticmethod
    def _validate_filter_file(file_reader):
        column_header = 0
        file_reader.open()
        for line in file_reader.read_lines():
            if line.startswith("chrom\tposition"):
                column_header = line
                break
        file_reader.close()

        if column_header:
            return file_reader

    @staticmethod
    def _is_varscan_vcf(file_reader):
        if file_reader.file_name.endswith(".vcf"):
            vcf_reader = vcf.VcfReader(file_reader)
            return "##source=VarScan2" in vcf_reader.metaheaders
        return False

#TODO: (cgates) Add check of header line - extract constant from HCTag?
    def _is_varscan_hc_file(self, file_reader):
        return self.hc_file_pattern.search(file_reader.file_name)

    @staticmethod
    def _raise_invalid_filter_exception(invalid_filter_files):
        if invalid_filter_files[5:]:
            omitted_files = ("...({} file(s) omitted)")\
                            .format(len(invalid_filter_files[5:]))
        else:
            omitted_files = ""
        first_five_fnames = [i.file_name for i in invalid_filter_files[:5]]
        raise utils.JQException("The [{}] input files [{}{}] match "
                                "high-confidence file names, but the "
                                "file header is invalid or missing. "
                                "Review inputs and try again.",
                                len(invalid_filter_files),
                                first_five_fnames,
                                omitted_files)

    @staticmethod
    def _validate_file_pairing(unpaired_hc_files, unpaired_vcf_files):
        if unpaired_hc_files:
            for unpaired_hc_file in unpaired_hc_files[:5]:
                msg = ("The VarScan high-confidence file [{}] has no matching "
                       "VCF file.")
                logger.error(msg, unpaired_hc_file.file_name)

            msg = ("[{}] VarScan high-confidence file(s) did not have a "
                   "matching VCF file. See log for more details. Review inputs "
                   "and try again.")
            raise utils.JQException(msg, len(unpaired_hc_files))

        if unpaired_vcf_files:
            for unpaired_vcf_file in unpaired_vcf_files[:5]:
                msg = ("The VarScan VCF file [{}] has no matching "
                       "high-confidence file.")
                logger.error(msg, unpaired_vcf_file.file_name)

            msg = ("[{}] VarScan VCF file(s) did not have a matching "
                   "high-confidence file. See log for more details. Review "
                   "inputs and try again.")
            raise utils.JQException(msg, len(unpaired_vcf_files))

    def _find_varscan_files(self, file_readers):
        unclaimed_set = set()
        prefix_file_readers = OrderedDict()
        filter_files = set()

        for file_reader in file_readers:
            if self._is_varscan_vcf(file_reader):
                prefix, _ = os.path.splitext(file_reader.file_name)
                prefix_file_readers[prefix] = file_reader
            elif self._is_varscan_hc_file(file_reader):
                filter_files.add(file_reader)
            else:
                unclaimed_set.add(file_reader)

        if not filter_files:
            if not self.hc_file_pattern.match(Varscan._DEFAULT_REGEX):
                msg = ("The VarScan high-confidence filename regex [{}] "
                       "didn't match any files in the input directory. "
                       "The beginning of the high-confidence filename must "
                       "exactly match a VCF filename up to the .vcf extention. "
                       "Review inputs/command options and try again.")
                raise utils.UsageError(msg, self.hc_file_pattern.pattern)

        return prefix_file_readers, filter_files, unclaimed_set

    @staticmethod
    def _split_prefix_by_patient(prefix_vcf_readers):
        prefix_by_patients = defaultdict(list)
        snp_indel = ["snp", "indel"]
        for prefix in prefix_vcf_readers:
            patient_names = [i for i in prefix.split(".") if i not in snp_indel]
            patient_name = ".".join(patient_names)
            prefix_by_patients[patient_name].append(prefix)

        return prefix_by_patients

    @staticmethod
    def _validate_vcf_readers(prefix_by_patients):
        number_of_files = set()
        for file_names in prefix_by_patients.values():
            if len(file_names) == 1:
                for file_name in file_names:
                    if re.search("snp", file_name):
                        msg = "VarScan VCF [{}] has no indel file."
                        logger.error(msg, file_name)
                    elif re.search("indel", file_name):
                        msg = "VarScan VCF [{}] has no snp file."
                        logger.error(msg, file_name)
            number_of_files.add(len(file_names))

        if len(number_of_files) > 1:
            msg = ("Some Varscan VCFs were missing either a snp or indel file. "
                   "Review inputs/command options and try again.")
            raise utils.JQException(msg)

    @staticmethod
    def _remove_paired_filters(vcf_dict, filter_dict):
        for filter_file in vcf_dict.values():
            try:
                filter_dict.pop(filter_file)
            except KeyError:
                pass
        return filter_dict

    @staticmethod
    def _dictionaries_to_tuples(vcf_dict, filter_dict):
        flipped_filter = OrderedDict((y, x) for x, y in filter_dict.items())
        pairs = vcf_dict.copy()
        pairs.update(flipped_filter)
        return [(k, v) for k, v in pairs.items()]

    def _pair_files(self, prefixes, filter_files):
        vcf_dict = OrderedDict()
        filter_dict = OrderedDict()
        invalid_filter_files = []

        for prefix, file_reader in prefixes.items():
            vcf_dict[file_reader] = None
            for filter_reader in filter_files:
                filter_dict[filter_reader] = None

                if filter_reader.file_name.startswith(prefix):
                    if self._validate_filter_file(filter_reader):
                        vcf_dict[file_reader] = filter_reader
                    else:
                        invalid_filter_files.append(filter_reader)

        if invalid_filter_files:
            self._raise_invalid_filter_exception(invalid_filter_files)
        filter_dict = self._remove_paired_filters(vcf_dict, filter_dict)

        return self._dictionaries_to_tuples(vcf_dict, filter_dict)

    def _validate_file_pairs(self, tuples):
        unpaired_vcf_files = []
        unpaired_hc_files = []

        hc_values = set([i[1] for i in tuples])
        if len(hc_values) != 1:
            for vcf_file_reader, hc_file_reader in tuples:
                if vcf_file_reader and not hc_file_reader:
                    unpaired_vcf_files.append(vcf_file_reader)
                if hc_file_reader and not vcf_file_reader:
                    unpaired_hc_files.append(hc_file_reader)

        self._validate_file_pairing(unpaired_hc_files, unpaired_vcf_files)

    @staticmethod
    def _create_vcf_readers(pair_tuples):
        vcf_readers = []
        for vcf_file_reader, hc_file_reader in pair_tuples:
            if vcf_file_reader and hc_file_reader:
                vcf_reader = _VarscanVcfReader(vcf.VcfReader(vcf_file_reader),
                                               hc_file_reader)
            elif vcf_file_reader and not hc_file_reader:
                vcf_reader = _VarscanVcfReader(vcf.VcfReader(vcf_file_reader))
            vcf_readers.append(vcf_reader)

        return vcf_readers

#pylint: disable=too-many-locals
    def claim(self, file_readers):
        """Recognizes and claims MuTect VCFs form the set of all input VCFs.

        Each defined caller has a chance to evaluate and claim all the incoming
        files as something that it can process. Since VarScan can claim
        high-confidence files as well, this process is significantly more
        complex than for other callers.

        Args:
            file_readers: the collection of currently unclaimed files

        Returns:
            A tuple of unclaimed readers and MuTectVcfReaders.
        """

        (prefix_to_readers,
         filter_files,
         unclaimed_set) = self._find_varscan_files(file_readers)

        prefix_by_patients = self._split_prefix_by_patient(prefix_to_readers)
        self._validate_vcf_readers(prefix_by_patients)
        vcf_filter_tuples = self._pair_files(prefix_to_readers, filter_files)
        self._validate_file_pairs(vcf_filter_tuples)
        vcf_readers = self._create_vcf_readers(vcf_filter_tuples)

        return list(unclaimed_set), vcf_readers

class _VarscanVcfReader(object):
    """Adapter that presents a VarScan VCF as a VcfReader.

    This follows the VcfReader interface, delegating calls to the base
    VcfReader, adjusting metaheaders and individual
    variants as appropriate.

    See VcfReader for more info.
    """
    def __init__(self, vcf_reader, som_hc_file_reader=None):
        self._vcf_reader = vcf_reader
        self._som_hc_file_reader = som_hc_file_reader
        self._caller = Varscan()
        self.tags = [common_tags.ReportedTag(JQ_VARSCAN_TAG),
                     common_tags.PassedTag(JQ_VARSCAN_TAG),
                     _AlleleFreqTag(),
                     _DepthTag(),
                     _SomaticTag()]

        if som_hc_file_reader:
            self.tags.insert(0, _HCTag(som_hc_file_reader))

    @property
    def file_name(self):
        return self._vcf_reader.file_name

    def _get_new_metaheaders(self):
        return [tag.metaheader for tag in self.tags]

    @property
    def caller_name(self):
        return self._caller.name

    def open(self):
        return self._vcf_reader.open()

    def close(self):
        return self._vcf_reader.close()

    @staticmethod
    def expected_file_format():
        return ["snp", "indel"]

    @property
    def metaheaders(self):
        new_metaheaders = list(self._vcf_reader.metaheaders)
        new_metaheaders.extend(self._get_new_metaheaders())
        caller_metaheader = "##jacquard.translate.caller={0}".\
                format(self._caller.name)
        new_metaheaders.append(caller_metaheader)

        return new_metaheaders

    @property
    def column_header(self):
        return self._vcf_reader.column_header

    def tagged_vcf_records(self):
        for vcf_record in self._vcf_reader.vcf_records():
            yield vcf_record

    def vcf_records(self):
        for vcf_record in self._vcf_reader.vcf_records():
            yield self._add_tags(vcf_record)

    def _add_tags(self, vcf_record):
        for tag in self.tags:
            tag.add_tag_values(vcf_record)
        return vcf_record



"""Merges a set of VCF files into a single VCF file.

* Merge assumes the incoming VCFs are aligned to the same set of contigs.
* Merge assumes incoming VCFs names follow this pattern:
    patientIdentifier.*.vcf
    Specifically, the first element of the VCF file name should be the patient
    name; for example you mihght have this:
        patientA.mutect.vcf, patientA.strelka.vcf, patientA.varscan.vcf,
        patientB.mutect.vcf, patientB.strelka.vcf, patientB.varscan.vcf,
        etc.
* Merge assumes the sample names (i.e. the VCF sample column headers, typically
    TUMOR and NORMAL) are consistent across the input VCFs. (The preceding
    Jacquard command "translate" ensures this is true.)
* Each incoming VCF record is joined with other VCF records that share the same
    "coordinate", where coordinate is the (chrom, pos, ref, and alt).
* Calls for the same patient/sample are joined into a single column. Each output
    column is the patient name (prefix of the file name) plus the sample name
    from the column header.
* By default, the merged file will have as many records as the distinct set of
    (chrom, pos, ref, alt) across all input files.
    Running Merge with "--include_variants" or "--include_loci" will reduce the
    number of records in the merged file to include only those which conform
    with the desired flag.
* Each variant record will have the minimal set of incoming format tags for
    that variant (i.e. the list of format tags is specific to each record).
* Incoming QUAL and INFO fields are ignored.
* By default, merge only includes "Jacquard" FORMAT tags, but other tags can be
    included through and optional a command line arg.
"""
from __future__ import print_function, absolute_import, division

import argparse
from collections import defaultdict, OrderedDict
import glob
import os
import re

import natsort

import jacquard.utils.logger as logger
import jacquard.utils.utils as utils
from jacquard.utils.vcf import FileWriter
import jacquard.utils.vcf as vcf


_DEFAULT_INCLUDED_FORMAT_TAGS = ["JQ_.*"]
_MULT_ALT_TAG = "JQ_MULT_ALT_LOCUS"
_JQ_SOMATIC_TAG = "HC_SOM"

_MULT_ALT_HEADER = ('##INFO=<ID={},Number=0,Type=Flag,'
                    'Description="More than one alt allele was seen at this '
                    'locus.">').format(_MULT_ALT_TAG)
_FILE_FORMAT = ["##fileformat=VCFv4.1"]
_FILE_OUTPUT_SUFFIX = "merged"

class _BufferedReader(object):
    """A look-ahead reader that expedites merging.

    Accepts an iterator and returns element (advancing current element) if
    requested element equals next element in collection and None otherwise.
    Never returns StopIteration so cannot be used as iteration control-flow.

    This behavior is useful only because VcfRecord equality is based on their
    coordinate (chrom, pos, ref, alt), so by iterating over a list of
    coordinates, you can either park on your next record or return the
    current record and advance the reader to the next.

    Each incoming VcfReader is wrapped in a a BufferedReader and readers are
    advanced to the next reader when their current coordinate is requested.
    This approach avoids reading all VCFs into memory, but does require a file
    handle for each VCF you are merging and also requires that VcfReaders are
    sorted identically.

    Stylistic note:
    This class must capture the state of the incoming iterator and provide
    modified behavior based on data in that iterator. A small class works ok,
    but suspect there may be a more pythonic way to curry iterator in a
    partial function. Uncertain if that would be clearer/simpler. [cgates]
    """
    #pylint: disable=too-few-public-methods
    def __init__(self, iterator):
        self._iterator = iterator
        try:
            self._current_element = next(self._iterator)
        except StopIteration:
            self._current_element = self._iterator

    def next_if_equals(self, requested_element):
        result = None
        if requested_element == self._current_element:
            result = self._current_element
            self._current_element = self._get_next()
        return result

    def _get_next(self):
        try:
            return next(self._iterator)
        except StopIteration:
            return None


class _Filter(object):
    """Interprets command-line args to initialize variant and locus filters
    """
    #pylint: disable=too-few-public-methods
    def __init__(self, args):
        self._cell_filters = {"all" : _Filter._include_cell_if_all,
                              "valid": _Filter._include_cell_if_valid,
                              "passed" : _Filter._include_cell_if_passed,
                              "somatic" : _Filter._include_cell_if_somatic}
        self._row_filters = {"all": _Filter._include_row_if_all,
                             "all_passed": _Filter._include_row_if_all_passed,
                             "at_least_one_passed": _Filter.\
                                                    _include_row_if_any_passed,
                             "all_somatic": _Filter._include_row_if_all_somatic,
                             "at_least_one_somatic": _Filter.\
                                                    _include_row_if_any_somatic}
        self._args = args
        self._cell_filter_strategy = self._cell_filters[args.include_cells]
        self._row_filter_strategy = self._row_filters[args.include_rows]

        self.row_count = 0
        self.rows_excluded = 0
        self.cell_count = 0
        self.cells_excluded = 0

        self.excluded_breakdown = defaultdict(int)

    def include_cell(self, vcf_record):
        self.cell_count += 1
        include = self._cell_filter_strategy(self, vcf_record)
        if not include:
            self.cells_excluded += 1
        return include

    def include_row(self, vcf_records):
        self.row_count += 1
        include = self._row_filter_strategy(vcf_records)
        if not include:
            self.rows_excluded += 1
        return include

    def log_statistics(self):
        cell_msg = "{}% ({}) cells were excluded because (--include_cells={})"
        logger.info(cell_msg,
                    int(round(100 * self.cells_excluded / self.cell_count, 0)),
                    self.cells_excluded,
                    self._args.include_cells)

        row_msg = "{}% ({}) rows were excluded because (--include_rows={})"
        logger.info(row_msg,
                    int(round(100 * self.rows_excluded / self.row_count, 0)),
                    self.rows_excluded,
                    self._args.include_rows)

        for key, count in self.excluded_breakdown.items():
            msg = "{} cells were excluded with [{}]"
            logger.debug(msg, count, key)

    #TODO: (cgates/jebene): Should this be part of VcfRecord?
    @staticmethod
    def _is_somatic(record):
        for sample in record.sample_tag_values:
            for tag in record.sample_tag_values[sample]:
                is_somatic = record.sample_tag_values[sample][tag] == "1"
                if re.search(_JQ_SOMATIC_TAG, tag) and is_somatic:
                    return True
        return False

    def _include_cell_if_all(self, dummy):
        #pylint: disable=no-self-use
        return True

    def _include_cell_if_somatic(self, record):
        include = _Filter._is_somatic(record)
        if not include:
            self.excluded_breakdown["not somatic"] += 1
        return include

    def _include_cell_if_valid(self, record):
        include ="JQ_EXCLUDE" not in record.filter
        if not include:
            self.excluded_breakdown[record.filter] += 1
        return include

    def _include_cell_if_passed(self, record):
        include = record.filter == "PASS"
        if not include:
            self.excluded_breakdown[record.filter] += 1
        return include

    @staticmethod
    def _include_row_if_all(dummy):
        return True

    @staticmethod
    def _include_row_if_all_passed(records):
        for record in records:
            if record.filter != "PASS":
                return False
        return True

    @staticmethod
    def _include_row_if_any_passed(records):
        for record in records:
            if record.filter == "PASS":
                return True
        return False

    @staticmethod
    def _include_row_if_all_somatic(records):
        for record in records:
            if not _Filter._is_somatic(record):
                return False
        return True

    @staticmethod
    def _include_row_if_any_somatic(records):
        for record in records:
            if _Filter._is_somatic(record):
                return True
        return False

def _build_format_tags(format_tag_regex, vcf_readers):
    retained_tags = set()
    regexes_used = set()
    for vcf_reader in vcf_readers:
        for tag_regex in format_tag_regex:
            for tag in vcf_reader.format_metaheaders:
                if re.match(tag_regex + "$", tag):
                    retained_tags.add(tag)
                    regexes_used.add(tag_regex)

    if len(retained_tags) == 0:
        msg = ("The specified format tag regex [{}] would exclude all format "
               "tags. Review inputs/usage and try again")
        raise utils.JQException(msg, format_tag_regex)

    unused_regexes = set(format_tag_regex).difference(regexes_used)
    if unused_regexes:
        for unused_regex in unused_regexes:
            msg = ("In the specified list of regexes {}, the regex [{}] does "
                   "not match any format tags; this expression may be "
                   "irrelevant.")
            logger.warning(msg, format_tag_regex, unused_regex)

    return sorted(list(retained_tags))

def _compile_metaheaders(incoming_headers,
                         vcf_readers,
                         all_sample_names,
                         contigs_to_keep,
                         format_tags_to_keep,
                         info_tags_to_keep):
    #pylint: disable=too-many-arguments
    ordered_metaheaders = list(incoming_headers)
    all_info_metaheaders = {}
    all_format_metaheaders = {}
    all_contig_metaheaders = {}

    for vcf_reader in vcf_readers:
        all_info_metaheaders.update(vcf_reader.info_metaheaders)
        all_format_metaheaders.update(vcf_reader.format_metaheaders)
        all_contig_metaheaders.update(vcf_reader.contig_metaheaders)

    all_info_metaheaders[_MULT_ALT_TAG] = _MULT_ALT_HEADER

    if all_contig_metaheaders:
        for contig_id in contigs_to_keep:
            ordered_metaheaders.append(all_contig_metaheaders[contig_id])
    for tag in info_tags_to_keep:
        ordered_metaheaders.append(all_info_metaheaders[tag])
    for tag in format_tags_to_keep:
        ordered_metaheaders.append(all_format_metaheaders[tag])

    sorted_metaheaders = utils.sort_metaheaders(ordered_metaheaders)

    column_header = vcf_readers[0].column_header.split("\t")[0:9]
    column_header.extend(all_sample_names)
    sorted_metaheaders.append("\t".join(column_header))

    return sorted_metaheaders

def _write_metaheaders(file_writer, all_headers):
    file_writer.write("\n".join(all_headers) + "\n")

def _create_vcf_readers(file_readers):
    vcf_readers = []
    for file_reader in file_readers:
        vcf_reader = vcf.VcfReader(file_reader)
        vcf_readers.append(vcf_reader)

    return vcf_readers

def _create_buffered_readers(vcf_readers):
    buffered_readers = []
    for vcf_reader in vcf_readers:
        vcf_reader.open()
        records = vcf_reader.vcf_records(qualified=True)
        buffered_readers.append(_BufferedReader(records))

    return buffered_readers

def _build_coordinates(vcf_readers):
    coordinate_set = set()
    mult_alts = defaultdict(set)

    for i, vcf_reader in enumerate(vcf_readers):
        logger.info("Reading [{}] ({}/{})",
                    os.path.basename(vcf_reader.file_name),
                    i + 1,
                    len(vcf_readers))
        try:
            vcf_reader.open()

            for vcf_record in vcf_reader.vcf_records():
                coordinate_set.add(vcf_record.get_empty_record())
                ref_alt = vcf_record.ref, vcf_record.alt
                locus = vcf_record.chrom, vcf_record.pos
                mult_alts[locus].add(ref_alt)
        finally:
            vcf_reader.close()

    for vcf_record in coordinate_set:
        ref_alts_for_this_locus = mult_alts[vcf_record.chrom, vcf_record.pos]
        inferred_mult_alt = len(ref_alts_for_this_locus) > 1
        explicit_mult_alt = "," in next(iter(ref_alts_for_this_locus))[1]

        if inferred_mult_alt or explicit_mult_alt:
            vcf_record.add_info_field(_MULT_ALT_TAG)

    return sorted(list(coordinate_set))

def _write_headers(reader, file_writer):
    headers = reader.metaheaders
    headers.append(reader.column_header)

    file_writer.write("\n".join(headers) + "\n")

def _sort_vcf(reader, sorted_dir):
    vcf_records = []
    reader.open()
    for vcf_record in reader.vcf_records():
        vcf_records.append(vcf_record)

    reader.close()
    vcf_records.sort()
    writer = FileWriter(os.path.join(sorted_dir,
                                     reader.file_name))
    writer.open()
    writer.write("\n".join(reader.metaheaders) + "\n")
    writer.write(reader.column_header + "\n")
    for vcf_record in vcf_records:
        writer.write(vcf_record.text())

    writer.close()
    reader = vcf.VcfReader(vcf.FileReader(writer.output_filepath))
    return reader

def _get_unsorted_readers(vcf_readers):
    unsorted_readers = []
    for i, reader in enumerate(vcf_readers):
        logger.info("Checking sort order of [{}] ({}/{})",
                    reader.file_name,
                    i+1,
                    len(vcf_readers)
                    )
        previous_record = None
        reader.open()
        for vcf_record in reader.vcf_records():
            if previous_record and vcf_record < previous_record:
                logger.debug("VCF file:chrom:pos [{}:{}:{}] is out of order"
                             .format(reader.file_name,
                                     vcf_record.chrom,
                                     vcf_record.pos))
                unsorted_readers.append(reader)
                break
            else:
                previous_record = vcf_record
        reader.close()
    return unsorted_readers

def _sort_readers(vcf_readers, output_path):
    unsorted_readers = _get_unsorted_readers(vcf_readers)
    sorted_readers = []
    unsorted_count = 0
    if unsorted_readers:
        sorted_dir = os.path.join(os.path.dirname(output_path), "tmp")
        os.makedirs(sorted_dir)

    for reader in vcf_readers:
        if reader in unsorted_readers:
            unsorted_count += 1
            logger.info("Sorting vcf [{}] ({}/{})",
                        reader.file_name,
                        unsorted_count,
                        len(unsorted_readers))
            reader = _sort_vcf(reader, sorted_dir)
        sorted_readers.append(reader)
    return sorted_readers


def _build_merged_record(coordinate,
                         vcf_records,
                         all_sample_names,
                         tags_to_keep):

    all_tags = set()
    sparse_matrix = {}

    for record in vcf_records:
        for sample, tags in record.sample_tag_values.items():
            if sample not in sparse_matrix:
                sparse_matrix[sample] = {}
            for tag, value in list(tags.items()):
                if tag in tags_to_keep:
                    all_tags.add(tag)
                    sparse_matrix[sample][tag] = value

    full_matrix = OrderedDict()
    for sample in all_sample_names:
        full_matrix[sample] = OrderedDict()
        for tag in sorted(all_tags):
            try:
                full_matrix[sample][tag] = sparse_matrix[sample][tag]
            except KeyError:
                full_matrix[sample][tag] = "."

    merged_record = vcf.VcfRecord(coordinate.chrom,
                                  coordinate.pos,
                                  coordinate.ref,
                                  coordinate.alt,
                                  coordinate.vcf_id,
                                  coordinate.qual,
                                  coordinate.filter,
                                  coordinate.info,
                                  sample_tag_values=full_matrix)

    return merged_record

def _pull_matching_records(filter_strategy, coordinate, buffered_readers):
    vcf_records = []
    for buffered_reader in buffered_readers:
        record = buffered_reader.next_if_equals(coordinate)
        if record and filter_strategy.include_cell(record):
            vcf_records.append(record)
    if filter_strategy.include_row(vcf_records):
        return vcf_records
    else:
        return []

def _build_sample_list(vcf_readers):
    all_sample_names = set()
    patient_to_file = defaultdict(list)

    for vcf_reader in vcf_readers:
        for sample_name in vcf_reader.qualified_sample_names:
            all_sample_names.add(sample_name)
            patient_to_file[sample_name].append(vcf_reader.file_name)
    sorted_sample_names = natsort.natsorted(all_sample_names)

    sorted_patient_to_file = OrderedDict()
    for sample_name in sorted_sample_names:
        sorted_patient_to_file[sample_name] = patient_to_file[sample_name]

    merge_metaheaders = _build_merge_metaheaders(sorted_patient_to_file)

    return sorted_sample_names, merge_metaheaders

def _build_info_tags(coordinates):
    all_info_tags = set()
    for record in coordinates:
        all_info_tags.update(record.info_dict.keys())
    ordered_tags = sorted(all_info_tags)

    return ordered_tags

def _build_contigs(coordinates):
    all_contigs = set()
    for record in coordinates:
        all_contigs.add(record.chrom)
    ordered_contigs = natsort.natsorted(all_contigs)

    return ordered_contigs

def _build_merge_metaheaders(patient_to_file):
    metaheaders = []
    metaheader_fmt = "##jacquard.merge.sample=<Column={0},Name={1},Source={2}>"
    for i, patient in enumerate(patient_to_file):
        filenames_string = "|".join(patient_to_file[patient])
        metaheader = metaheader_fmt.format(i + 1, patient, filenames_string)
        metaheaders.append(metaheader)

    return metaheaders

def _build_writers_to_readers(vcf_readers, output_path):
    writers_to_readers = {}
    basename, extension = os.path.splitext(output_path)
    new_filename = basename + ".merged" + extension

    file_writer = vcf.FileWriter(new_filename)
    writers_to_readers[file_writer] = vcf_readers

    return writers_to_readers

def _get_readers_per_patient(file_readers):
    readers_per_patient = defaultdict(list)
    for file_reader in file_readers:
        file_reader.open()
        patient = file_reader.file_name.split(".")[0]
        for line in file_reader.read_lines():
            caller_meta_header = "##jacquard.translate.caller"
            if line.startswith(caller_meta_header):
                readers_per_patient[patient].append(line.split("=")[1])
        file_reader.close()

    return OrderedDict(sorted(readers_per_patient.items()))

def _validate_consistent_samples(file_readers):
    readers_per_patient = _get_readers_per_patient(file_readers)

    all_callers = set()
    for callers in readers_per_patient.values():
        all_callers.update(callers)
    warning = 0
    for patient, callers in readers_per_patient.items():
        missing_callers = set(all_callers).difference(set(callers))
        if missing_callers:
            warning = 1
            msg = "Sample [{}] is missing VCF(s): {}"
            logger.warning(msg,
                           patient,
                           sorted(list(missing_callers)))
    if warning:
        msg = "Some samples appear to be missing VCF(s)"
        logger.warning(msg)

def _merge_records(vcf_readers,
                   coordinates,
                   filter_strategy,
                   all_sample_names,
                   format_tags_to_keep,
                   file_writer):
    #pylint: disable=too-many-arguments
    row_count = 0
    next_breakpoint = 0
    buffered_readers = _create_buffered_readers(vcf_readers)

    for coordinate in coordinates:
        vcf_records = _pull_matching_records(filter_strategy,
                                             coordinate,
                                             buffered_readers)
        if vcf_records:
            merged_record = _build_merged_record(coordinate,
                                                 vcf_records,
                                                 all_sample_names,
                                                 format_tags_to_keep)
            file_writer.write(merged_record.text())

        row_count += 1
        progress = 100 * row_count / len(coordinates)
        if progress > next_breakpoint:
            logger.info("Merging: {} rows processed ({}%)",
                        row_count,
                        next_breakpoint)
            next_breakpoint = 10 * int(progress/10) + 10

    logger.info("Merge complete: {} rows processed (100%)", row_count)
    filter_strategy.log_statistics()

def add_subparser(subparser):
    #pylint: disable=line-too-long
    parser = subparser.add_parser("merge", formatter_class=argparse.RawTextHelpFormatter, help="Accepts a directory of VCFs and returns a single merged VCF file.")
    parser.add_argument("input", help="Path to directory containing VCFs. Other file types ignored")
    parser.add_argument("output", help="Path to output variant-level VCF file")
    parser.add_argument("-v", "--verbose", action='store_true')
    parser.add_argument("--force", action='store_true', help="Overwrite contents of output directory")
    parser.add_argument("--include_format_tags", dest='tags', help="Comma-separated list of regexs for format tags to include in output. Defaults to all JQ tags.")
    parser.add_argument("--log_file", help="Log file destination")
    parser.add_argument("--include_cells",
                        choices=["all", "valid", "passed", "somatic"],
                        default="valid",
                        help=("all: Include all variants\n"
                              "valid: Only include valid variants\n"
                              "passed: Only include variants which passed their respective filter\n"
                              "somatic: Only include somatic variants"),
                        metavar=["all", "valid", "passed", "somatic"])
#                         metavar="")
    parser.add_argument("--include_rows",
                        choices=["all", "at_least_one_passed", "all_passed", "at_least_one_somatic", "all_somatic"],
                        default="at_least_one_somatic",
                        help=("all: Include all variants at loci\n"
                              #pylint: disable=line-too-long
                              "at_least_one_passed: Include all variants at loci where at least one variant passed\n"
                              "all_passed: Include all variants at loci where all variants passed\n"
                              "at_least_one_somatic: Include all variants at loci where at least one variant was somatic\n"
                              "all_somatic: Include all variants at loci where all variants were somatic"),
                        metavar=["all", "at_least_one_passed", "at_least_one_somatic", "all_somatic"])

def _predict_output(args):
    desired_output_files = set([os.path.basename(args.output)])

    return desired_output_files

def report_prediction(args):
    return _predict_output(args)

def get_required_input_output_types():
    return ("directory", "file")

#TODO (cgates): Validate should actually validate
def validate_args(dummy):
    pass

def execute(args, execution_context):
    input_path = os.path.abspath(args.input)
    output_path = os.path.abspath(args.output)
    filter_strategy = _Filter(args)
    if args.tags:
        format_tag_regex = args.tags.split(",")
    else:
        format_tag_regex = _DEFAULT_INCLUDED_FORMAT_TAGS

    input_files = sorted(glob.glob(os.path.join(input_path, "*.vcf")))
    file_readers = [vcf.FileReader(i) for i in input_files]
    _validate_consistent_samples(file_readers)

    try:
        file_writer = vcf.FileWriter(output_path)
        file_writer.open()

#TODO: jebene: _build_format_tags, _built_sample_list, _build_info_tags,
#and _build_contigs behave differently. It seems like we could make the
#signatures of these methods more similar or even combine some methods to
#reduce excess iterations over the coordinates/vcf_readers
        vcf_readers = _create_vcf_readers(file_readers)
        format_tags_to_keep = _build_format_tags(format_tag_regex, vcf_readers)
        vcf_readers = _sort_readers(vcf_readers, output_path)
        all_sample_names, merge_metaheaders = _build_sample_list(vcf_readers)
        coordinates = _build_coordinates(vcf_readers)
        info_tags_to_keep = _build_info_tags(coordinates)
        contigs_to_keep = _build_contigs(coordinates)
        incoming_headers = _FILE_FORMAT + execution_context + merge_metaheaders
        headers = _compile_metaheaders(incoming_headers,
                                       vcf_readers,
                                       all_sample_names,
                                       contigs_to_keep,
                                       format_tags_to_keep,
                                       info_tags_to_keep)

        _write_metaheaders(file_writer, headers)

        _merge_records(vcf_readers,
                       coordinates,
                       filter_strategy,
                       all_sample_names,
                       format_tags_to_keep,
                       file_writer)
    finally:
        for vcf_reader in vcf_readers:
            vcf_reader.close()
        file_writer.close()



"""Functions and classes common to multiple modules.

"""
#By placing common functions here instead of in the jacquard module, we avoid
#circular dependencies. (e.g. jacquard depends on translate depends on jacquard)
from __future__ import print_function, absolute_import, division

import natsort


#TODO (cgates): Why does this need a string? Seems like it should take a number?
def round_two_digits(val):
    try:
        if len(val.split(".")[1]) > 2:
            return "{0:.2f}".format(float(val))
    except IndexError:
        return val
    return val

def sort_metaheaders(metaheaders):
    metaheader_dict = {"##fileformat": 1,
                       "##jacquard": 2,
                       "##contig": 3,
                       "##ALT": 4,
                       "##FILTER": 5,
                       "##INFO": 6,
                       "##FORMAT": 7,
                       "#CHROM": 8}

    for metaheader in metaheaders:
        prefix = metaheader.split("=")[0]
        if prefix not in metaheader_dict:
            metaheader_dict[prefix] = 2

    nat_sorted_metaheaders = natsort.natsorted(metaheaders)

    return sorted(nat_sorted_metaheaders,
                  key=lambda metaheader: metaheader_dict[metaheader\
                                                         .split("=")[0]])

class JQException(Exception):
    """Base class for all run-time exceptions in this module."""
    def __init__(self, msg, *args):
        #pylint: disable=star-args
        error_msg = msg.format(*[str(i) for i in args])
        super(JQException, self).__init__(error_msg)


class UsageError(JQException):
    """Raised for malformed command or invalid arguments."""
    def __init__(self, msg, *args):
        super(UsageError, self).__init__(msg, *args)




"""Delegates to third-party variant callers.

To keep things simple, fair, and encapsulated route any requests to variant
callers through this module; likewise, don't directly reference individual
variant callers outside of this module.
"""
from __future__ import print_function, absolute_import, division

import jacquard.utils.logger as logger
import jacquard.utils.utils as utils
import jacquard.variant_caller_transforms.mutect as mutect
import jacquard.variant_caller_transforms.strelka as strelka
import jacquard.variant_caller_transforms.varscan as varscan


SUPPORTED_CALLER_VERSIONS = {"VarScan": varscan.VERSION,
                             "MuTect": mutect.VERSION,
                             "Strelka": strelka.VERSION}

class VariantCallerFactory(object):
    def __init__(self, args=None):
        self._callers = [varscan.Varscan(args),
                         strelka.Strelka(),
                         mutect.Mutect()]

#TODO: (cgates): Filter uses this, but only for logging; adjust filter and drop
# method. Then consider renaming the module.
    def get_caller(self, metaheaders, column_header, name):
        for caller in self._callers:
            if caller.validate_input_file(metaheaders, column_header):
                logger.debug("VCF [{}] recognized by caller [{}]",
                             name,
                             caller.name)
                return caller
        raise utils.JQException(("VCF [{}] was not in the set of "
                                 "recognized callers.").format(name))

    def claim(self, unclaimed_file_readers):
        """Allows each caller to claim incoming files as they are recognized.

        Args:
            unclaimed_file_readers: Usually, all files in the input dir.

        Returns:
            A tuple of unclaimed file readers and claimed VcfReaders. The
            presence of any unclaimed file readers could indicate stray files
            in the input dir.
        """
        claimed_vcf_readers = []
        for caller in self._callers:
            (unclaimed_file_readers,
             translated_vcf_readers) = caller.claim(unclaimed_file_readers)
            claimed_vcf_readers.extend(translated_vcf_readers)

        return unclaimed_file_readers, claimed_vcf_readers



"""Adds summary tags/fields to a merged VCF file.

Collaborates with two summary "callers" to add INFO and FORMAT tags to each
variant record based on the presence of previously translated tags.
"""
from __future__ import print_function, absolute_import, division

import argparse
import os

import jacquard.utils.logger as logger
import jacquard.utils.utils as utils
import jacquard.utils.summarize_rollup_transform as summarize_caller
import jacquard.utils.summarize_zscore_transform as zscore_caller
import jacquard.utils.vcf as vcf


def _write_metaheaders(caller,
                       vcf_reader,
                       file_writer,
                       execution_context=None,
                       new_meta_headers=None):

    new_headers = list(vcf_reader.metaheaders)

    if execution_context:
        new_headers.extend(execution_context)
        new_headers.extend(caller.get_metaheaders())
    if new_meta_headers:
        new_headers.append(new_meta_headers)

    sorted_metaheaders = utils.sort_metaheaders(new_headers)
    sorted_metaheaders.append(vcf_reader.column_header)

    file_writer.write("\n".join(sorted_metaheaders) +"\n")

def _write_to_tmp_file(caller, vcf_reader, tmp_writer):
    vcf_reader.open()
    tmp_writer.open()

    try:
        _write_metaheaders(caller, vcf_reader, tmp_writer)
        logger.info("Adding summary tags for [{}]", vcf_reader.file_name)
        _add_tags(caller, vcf_reader, tmp_writer)

    finally:
        vcf_reader.close()
        tmp_writer.close()


def _write_zscores(caller,
                   metaheaders,
                   vcf_reader,
                   file_writer):

#TODO: (jebene) make zscores and tmp file follow the same pattern when writing
    try:
        file_writer.open()
        headers = list(metaheaders)
        headers.extend(vcf_reader.metaheaders)
        headers.extend(caller.metaheaders)
        sorted_metaheaders = utils.sort_metaheaders(headers)
        sorted_metaheaders.append(vcf_reader.column_header)
        file_writer.write("\n".join(sorted_metaheaders) +"\n")

        vcf_reader.open()
        for vcf_record in vcf_reader.vcf_records():
            line = caller.add_tags(vcf_record)
            file_writer.write(line)
    finally:
        vcf_reader.close()
        file_writer.close()

def _add_tags(caller, vcf_reader, file_writer):
    for vcf_record in vcf_reader.vcf_records():
        caller.add_tags(vcf_record)
        file_writer.write(vcf_record.text())

def add_subparser(subparser):
    # pylint: disable=line-too-long
    parser = subparser.add_parser("summarize", formatter_class=argparse.RawTextHelpFormatter, help="Accepts a Jacquard-merged VCF file and creates a new file, adding summary fields.")
    parser.add_argument("input", help="Path to Jacquard-merged VCF (or any VCF with Jacquard tags; e.g. JQ_SOM_MT)")
    parser.add_argument("output", help="Path to output VCf")
    parser.add_argument("-v", "--verbose", action='store_true')
    parser.add_argument("--force", action='store_true', help="Overwrite contents of output directory")
    parser.add_argument("--log_file", help="Log file destination")

def report_prediction(args):
    return set([os.path.basename(args.output)])

def get_required_input_output_types():
    return ("file", "file")

#TODO (cgates): Validate should actually validate
def validate_args(dummy):
    pass

def execute(args, execution_context):
    input_file = os.path.abspath(args.input)
    output = os.path.abspath(args.output)

    summary_caller = summarize_caller.SummarizeCaller()

    vcf_reader = vcf.VcfReader(vcf.FileReader(input_file))
    tmp_output_file = output + ".tmp"
    tmp_writer = vcf.FileWriter(tmp_output_file)

    _write_to_tmp_file(summary_caller, vcf_reader, tmp_writer)

    tmp_reader = vcf.VcfReader(vcf.FileReader(tmp_output_file))
    file_writer = vcf.FileWriter(output)

    logger.info("Calculating zscores")
    caller = zscore_caller.ZScoreCaller(tmp_reader)
    metaheaders = execution_context + summary_caller.get_metaheaders()
    _write_zscores(caller, metaheaders, tmp_reader, file_writer)

    os.remove(tmp_output_file)
